/*****************************************************************************
 *
 * tsusbhid.c -- TS2005F-USB Touchscreen Driver for Linux Kernel 2.4.x
 *
*****************************************************************************/

#define TSUSB_MAJOR 234

#include <linux/config.h>
#include <linux/module.h>
#include <linux/string.h>
#include <linux/poll.h>
#include <linux/slab.h>
#include <linux/smp_lock.h>
#include <linux/usb.h>
#include <linux/input.h>
#include <linux/devfs_fs_kernel.h>

#include "tsusbhid.h"

#define DRIVER_VERCODE		TSUSB_DRIVER_VERCODE
#define DRIVER_VERSION		TSUSB_DRIVER_VERSION
#define DRIVER_AUTHOR		"CometrueWorks"
#define DRIVER_DESC		"TS2005F-USB Touchscreen Driver for Linux Kernel 2.4.x"
#define DRIVER_LICENSE		"GPL"
#define DRIVER_NAME		"tsusbhid"
#define DEVICE_NAME		"TS2005F-USB"

#undef	TSUSB_WAKE_ALL
#ifdef	TSUSB_WAKE_ALL
#define	WAKE_UP(x)  wake_up(x);
#define	SLEEP_ON(x) sleep_on(x);
#else
#define	WAKE_UP(x)  wake_up_interruptible(x);
#define	SLEEP_ON(x) sleep_interruptible_on(x);
#endif

#define TSKI(arg...) printk(KERN_INFO    DRIVER_NAME ": " arg)
#define TSKE(arg...) printk(KERN_ERR     DRIVER_NAME ": " arg)
#define TSKW(arg...) printk(KERN_WARNING DRIVER_NAME ": " arg)
#define TSKP(arg...) printk(arg)

#define	TOUCH_MAJOR		TSUSB_MAJOR
#define	TOUCH_MAJOR_NAME	TSUSB_MAJOR_NAME
#define	TOUCH_MINOR_NAME	TSUSB_MINOR_NAME
#define	TOUCH_MINOR_BASE	TSUSB_MINOR_BASE
#define	TOUCH_MINORS		TSUSB_MINORS

#define	TOUCH_DEVFS_NAME_SIZE	32
#define	TOUCH_DEVFS_NAME	TOUCH_MAJOR_NAME "/" TOUCH_MINOR_NAME "%d"
#define	TOUCH_DEVFS_MODE	S_IFCHR | S_IRUGO | S_IWUSR

struct tsusb_hid_device *tsusb_hiddev_table[TOUCH_MINORS] = { NULL, /* ... */ };

struct tsusb_event_device *tsusb_evdev = NULL;

struct file_operations *touch_minors_fops_table[TOUCH_MINORS] = { NULL, /* ... */ };
spinlock_t              touch_minors_fops_lock = SPIN_LOCK_UNLOCKED;

devfs_handle_t touch_devfs_handle = NULL;
devfs_handle_t tsusb_devfs_handle[TOUCH_MINORS] = { NULL, /* ... */ };
unsigned int   touch_devfs_major = TOUCH_MAJOR;

static void tsusbhid_evdev_event( unsigned short int id, unsigned short int v1, unsigned short int v2, unsigned short int v3 )
{
	struct tsusb_event_device_list *list = tsusb_evdev->list;

	while (list) {
		list->buffer[list->head].id = id;
		list->buffer[list->head].v1 = v1;
		list->buffer[list->head].v2 = v2;
		list->buffer[list->head].v3 = v3;
		list->head = (list->head + 1) & (TSUSB_LIST_BUFFER_SIZE - 1);
		kill_fasync(&list->fasync, SIGIO, POLL_IN);
		list = list->next;
	}
	WAKE_UP(&tsusb_evdev->wait);
}

static void tsusbhid_irq(struct urb *urb)
{
	struct tsusb_hid_device *tshiddev = (struct tsusb_hid_device *) urb->context;
	struct tsusb_input_report *input = tshiddev->input_report;
	struct tsusb_hid_device_list *list = tshiddev->list;

	if (urb->status) {
		return;
	}

	tsusbhid_evdev_event(TSUSB_EVENT_INPUT_DATA, (tshiddev->devnum<<8|input->state), input->xdata, input->ydata );

	while (list) {
		list->buffer[list->head].state = input->state;
		list->buffer[list->head].xdata = input->xdata;
		list->buffer[list->head].ydata = input->ydata;
		list->head = (list->head + 1) & (TSUSB_LIST_BUFFER_SIZE - 1);
		kill_fasync(&list->fasync, SIGIO, POLL_IN);
		list = list->next;
	}

	WAKE_UP(&tshiddev->wait);

}

static int tsusbhid_set_feature (struct tsusb_hid_device *tshiddev)
{
	int ret;

	ret = usb_control_msg(
		tshiddev->usbdev,
		usb_sndctrlpipe( tshiddev->usbdev, 0 ),
		TSHID_REQ_SET_REPORT,
		TSHID_REQ_SET,
		TSHID_REQ_TYPE_ID(TSHID_REPORT_TYPE_FEATURE,0),
		0,
		&tshiddev->feature,
		sizeof(struct tsusb_feature_report),
		300
		);

	if ( ret < 0 ) {
	}
	else {
		if ( ret != sizeof(struct tsusb_feature_report) ) {
			ret = -EIO;
		}
		else {
			ret = 0;
		}

	}

	return ret;
}

static int tsusbhid_get_feature (struct tsusb_hid_device *tshiddev)
{
	int ret;

	ret = usb_control_msg(
		tshiddev->usbdev,
		usb_rcvctrlpipe( tshiddev->usbdev, 0 ),
		TSHID_REQ_GET_REPORT,
		TSHID_REQ_GET,
		TSHID_REQ_TYPE_ID(TSHID_REPORT_TYPE_FEATURE,0),
		0,
		&tshiddev->feature,
		sizeof(struct tsusb_feature_report),
		300
		);
	if ( ret < 0 ) {
	}
	else {
		if ( ret != sizeof(struct tsusb_feature_report) ) {
			ret = -EIO;
		}
		else {
			ret = 0;
		}
	}

	return ret;
}

static ssize_t tsusbhid_write(struct file * file, const char * buffer, size_t count, loff_t *ppos)
{
	int retval = 0;
	return retval;
}

static ssize_t tsusbhid_read (struct file * file, char * buffer, size_t count, loff_t *ppos)
{
	DECLARE_WAITQUEUE(wait, current);
	struct tsusb_hid_device_list *list = file->private_data;
	int retval = 0;

	if (list->head == list->tail) {

		add_wait_queue(&list->tshiddev->wait, &wait);

		current->state = TASK_INTERRUPTIBLE;

		while (list->head == list->tail) {

			if (!list->tshiddev->exist) {
				retval = -ENODEV;
				break;
			}
			if (file->f_flags & O_NONBLOCK) {
				retval = -EAGAIN;
				break;
			}
			if (signal_pending(current)) {
				retval = -ERESTARTSYS;
				break;
			}

			schedule();
		}

		current->state = TASK_RUNNING;
		remove_wait_queue(&list->tshiddev->wait, &wait);
	}

	if (retval) {
		return retval;
	}

	while (list->head != list->tail && retval + sizeof(struct tsusb_input_report) <= count) {
		if (copy_to_user(
			buffer + retval,
			list->buffer + list->tail,
			sizeof(struct tsusb_input_report))) {
			return -EFAULT;
		}
		list->tail = (list->tail + 1) & (TSUSB_LIST_BUFFER_SIZE - 1);
		retval += sizeof(struct tsusb_input_report);
	}

	return retval;

}

static int tsusbhid_open (struct inode *inode, struct file *file)
{
	int minor = MINOR (inode->i_rdev);
	struct tsusb_hid_device *tshiddev;
	struct tsusb_hid_device_list *list;

	if ( minor < TOUCH_MINOR_BASE || minor > TOUCH_MINOR_BASE+TOUCH_MINORS-1) {
		TSKE("OPEN failed! minor %d, out of range %d to %d\n",
		minor,
		TOUCH_MINOR_BASE,
		TOUCH_MINOR_BASE+TOUCH_MINORS-1);
		return -EIO;
	}

	tshiddev = tsusb_hiddev_table[minor - TOUCH_MINOR_BASE];

	if (!(list = kmalloc(sizeof(struct tsusb_hid_device_list), GFP_KERNEL))) {
		TSKE("OPEN failed! unable to alloc buffer\n");
		return -ENOMEM;
	}
	memset(list, 0, sizeof(struct tsusb_hid_device_list));

	list->tshiddev = tshiddev;
	list->next  = tshiddev->list;
	tshiddev->list = list;

	file->private_data = list;

	if (!list->tshiddev->open++) {
	}

	return 0;
}

static int tsusbhid_fasync(int fd, struct file *file, int on)
{
	int retval;
	struct tsusb_hid_device_list *list = file->private_data;

	retval = fasync_helper(fd, file, on, &list->fasync);

	return retval < 0 ? retval : 0;
}

static int tsusbhid_release (struct inode *inode, struct file *file)
{
	struct tsusb_hid_device_list *list = file->private_data;
	struct tsusb_hid_device_list **listptr;

	lock_kernel();

	listptr = &list->tshiddev->list;

	tsusbhid_fasync(-1, file, 0);

	while (*listptr && (*listptr != list)) {
		listptr = &((*listptr)->next);
	}

	*listptr = (*listptr)->next;

	if (!--list->tshiddev->open) {
	}

	kfree(list);

	unlock_kernel();

	return 0;

}

static unsigned int tsusbhid_poll(struct file *file, poll_table *wait)
{
	struct tsusb_hid_device_list *list = file->private_data;

	poll_wait(file, &list->tshiddev->wait, wait);

	if (list->head != list->tail) {
		return POLLIN | POLLRDNORM;
	}

	return 0;
}

static int tsusbhid_ioctl (struct inode *inode, struct file *file, unsigned int cmd, unsigned long arg)
{

	struct tsusb_hid_device_list *list = file->private_data;
	struct tsusb_hid_device *tshiddev = list->tshiddev;
	struct usb_device *usbdev = tshiddev->usbdev;

	int ret = 0;
	int err = 0;
	int version = DRIVER_VERCODE;

	if (tshiddev->remove_pending) {
		TSKE("IOCTL failed! Device remove is pending.\n");
		return -EIO;
	}

	if (!tshiddev->exist) {
		TSKE("IOCTL failed! Device removed.\n");
		return -EIDRM;
	}

	if (!tshiddev->usbdev) {
		TSKE("IOCTL failed! No such device.\n");
		return -ENODEV;
	}

	if (_IOC_TYPE(cmd) != TSHIDIOC_TYPE ) {
		TSKE("IOCTL failed! Bad command type.\n");
		return -ENOTTY;
	}

	if (_IOC_NR(cmd) > TSHIDIOC_MAX_NR ) {
		TSKE("IOCTL failed! Bad command number.\n");
		return -ENOTTY;
	}

	if (_IOC_DIR(cmd) & _IOC_READ)
		err = !access_ok(VERIFY_WRITE, (void *)arg, _IOC_SIZE(cmd));
        else if (_IOC_DIR(cmd) & _IOC_WRITE)
		err = !access_ok(VERIFY_READ, (void *)arg, _IOC_SIZE(cmd));

	if (err) {
		TSKE("IOCTL failed! Bad argument buffer.\n");
		return -EFAULT;
	}

	if (tshiddev->buzy) {
		TSKE("IOCTL failed! Device or resource busy.\n");
		return -EBUSY;
	}

	tshiddev->buzy = 1;

	switch (cmd) {

	case TSHIDIOC_GET_VERSION:
		ret = put_user (version, (unsigned int *) arg);
		break;

	case TSHIDIOC_GET_ID:
		if ((ret = put_user(usbdev->descriptor.idVendor,  ((tshidioc_id_type *) arg) + TSHIDIOC_ID_VENDOR   ))) break;
		if ((ret = put_user(usbdev->descriptor.idProduct, ((tshidioc_id_type *) arg) + TSHIDIOC_ID_PRODUCT  ))) break;
		if ((ret = put_user(usbdev->descriptor.bcdDevice, ((tshidioc_id_type *) arg) + TSHIDIOC_ID_REVISION ))) break;
		if ((ret = put_user(usbdev->bus->busnum,          ((tshidioc_id_type *) arg) + TSHIDIOC_ID_BUSNUM   ))) break;
		if ((ret = put_user(usbdev->devnum,               ((tshidioc_id_type *) arg) + TSHIDIOC_ID_DEVNUM   ))) break;
                break;

	case TSHIDIOC_GET_FEATURE:
		if (_IOC_SIZE(cmd) != sizeof(struct tsusb_feature_report)) {
			ret = -EFAULT;
			break;
		}
		ret = tsusbhid_get_feature (tshiddev);
		if ( ret == 0 ) {
			ret = copy_to_user ((void *) arg, &tshiddev->feature, sizeof(struct tsusb_feature_report));
		}
		break;

	case TSHIDIOC_SET_FEATURE:
		if (_IOC_SIZE(cmd) != sizeof(struct tsusb_feature_report)) {
			ret = -EFAULT;
			break;
		}
		ret = copy_from_user (&tshiddev->feature, (void *) arg, sizeof(struct tsusb_feature_report));
		if ( ret == 0 ) {
			ret = tsusbhid_set_feature (tshiddev);
		}
		break;

	case TSHIDIOC_SET_ENABLE:
		tsusbhid_evdev_event(TSUSB_EVENT_ENABLE,tshiddev->devnum,0,0);
		break;

	case TSHIDIOC_SET_DISABLE:
		tsusbhid_evdev_event(TSUSB_EVENT_DISABLE,tshiddev->devnum,0,0);
		break;

	default:
		TSKE("IOCTL failed! No ioctl command (0x%x).\n",cmd);
		ret = -ENOIOCTLCMD;
		break;
	}

	tshiddev->buzy = 0;

	return ret;
}

static int touch_open(struct inode * inode, struct file * file)
{
	return -ENODEV;
}

static ssize_t tsusbhid_evdev_write(struct file * file, const char * buffer, size_t count, loff_t *ppos)
{
	int retval = 0;

	return retval;
}

static ssize_t tsusbhid_evdev_read (struct file * file, char * buffer, size_t count, loff_t *ppos)
{
	DECLARE_WAITQUEUE(wait, current);

	struct tsusb_event_device_list *list = file->private_data;

	int retval = 0;

	if (list->head == list->tail) {

		add_wait_queue(&list->tsevdev->wait, &wait);

		current->state = TASK_INTERRUPTIBLE;

		while (list->head == list->tail) {

			if (file->f_flags & O_NONBLOCK) {
				retval = -EAGAIN;
				break;
			}
			if (signal_pending(current)) {
				retval = -ERESTARTSYS;
				break;
			}

			schedule();
		}

		current->state = TASK_RUNNING;
		remove_wait_queue(&list->tsevdev->wait, &wait);
	}

	if (retval) {
		return retval;
	}

	while (list->head != list->tail && retval + sizeof(struct tsusb_event_packet) <= count) {
		if (copy_to_user(
			buffer + retval,
			list->buffer + list->tail,
			sizeof(struct tsusb_event_packet))) {
			return -EFAULT;
		}
		list->tail = (list->tail + 1) & (TSUSB_LIST_BUFFER_SIZE - 1);
		retval += sizeof(struct tsusb_event_packet);
	}

	return retval;

}

static int tsusbhid_evdev_open (struct inode *inode, struct file *file)
{
	int minor = MINOR (inode->i_rdev);
	struct tsusb_event_device_list *list;
	struct tsusb_hid_device *tshiddev;
	int u;

	if ( minor != TOUCH_MINOR_BASE+TOUCH_MINORS ) {
		return -EIO;
	}

	if (!(list = kmalloc(sizeof(struct tsusb_event_device_list), GFP_KERNEL))) {
		return -ENOMEM;
	}
	memset(list, 0, sizeof(struct tsusb_event_device_list));

	list->tsevdev  = tsusb_evdev;
	list->next  = tsusb_evdev->list;
	tsusb_evdev->list = list;

	file->private_data = list;

	if (!list->tsevdev->open++) {
	}

	for (u = 0; u < TOUCH_MINORS && tsusb_hiddev_table[u]; u++) {

		tshiddev = tsusb_hiddev_table[u];
		if ( tshiddev->exist  ) {
			list->buffer[list->head].id  = TSUSB_EVENT_PLUG;
			list->buffer[list->head].v1 = tshiddev->devnum;
			list->buffer[list->head].v2 = 0;
			list->buffer[list->head].v3 = 0;
			list->head = (list->head + 1) & (TSUSB_LIST_BUFFER_SIZE - 1);
			kill_fasync(&list->fasync, SIGIO, POLL_IN);
		}
	}

	WAKE_UP(&tsusb_evdev->wait);

	return 0;
}

static int tsusbhid_evdev_fasync(int fd, struct file *file, int on)
{
	int retval;
	struct tsusb_event_device_list *list = file->private_data;

	retval = fasync_helper(fd, file, on, &list->fasync);

	return retval < 0 ? retval : 0;
}

static int tsusbhid_evdev_release (struct inode *inode, struct file *file)
{
	struct tsusb_event_device_list *list = file->private_data;
	struct tsusb_event_device_list **listptr;

	lock_kernel();

	listptr = &list->tsevdev->list;

	tsusbhid_evdev_fasync(-1, file, 0);

	while (*listptr && (*listptr != list)) {
		listptr = &((*listptr)->next);
	}

	*listptr = (*listptr)->next;

	if (!--list->tsevdev->open) {
	}

	kfree(list);

	unlock_kernel();

	return 0;

}

static unsigned int tsusbhid_evdev_poll(struct file *file, poll_table *wait)
{
	struct tsusb_event_device_list *list = file->private_data;

	poll_wait(file, &list->tsevdev->wait, wait);

	if (list->head != list->tail) {
		return POLLIN | POLLRDNORM;
	}

	return 0;
}

static int tsusbhid_evdev_ioctl (struct inode *inode, struct file *file, unsigned int cmd, unsigned long arg)
{
	struct tsusb_event_device_list *list = file->private_data;
	struct tsusb_event_device *tsevdev = list->tsevdev;

	int ret = 0;
	int err = 0;
	int version = DRIVER_VERCODE;

	if (_IOC_TYPE(cmd) != TSUSBIOC_TYPE ) {
		TSKE("IOCTL failed! Bad command type.\n");
		return -ENOTTY;
	}

	if (_IOC_NR(cmd) > TSUSBIOC_MAX_NR ) {
		TSKE("IOCTL failed! Bad command number.\n");
		return -ENOTTY;
	}

	if (_IOC_DIR(cmd) & _IOC_READ)
		err = !access_ok(VERIFY_WRITE, (void *)arg, _IOC_SIZE(cmd));
        else if (_IOC_DIR(cmd) & _IOC_WRITE)
		err = !access_ok(VERIFY_READ, (void *)arg, _IOC_SIZE(cmd));

	if (err) {
		TSKE("IOCTL failed! Bad argument buffer.\n");
		return -EFAULT;
	}

	if (tsevdev->buzy) {
		TSKE("IOCTL failed! Device or resource busy.\n");
		return -EBUSY;
	}

	tsevdev->buzy = 1;

	switch (cmd) {

	case TSUSBIOC_GET_VERSION:
		ret = put_user (version, (unsigned int *) arg);
		break;
	default:
		TSKE("IOCTL failed! No ioctl command (0x%x).\n",cmd);
		ret = -ENOIOCTLCMD;
	}

	tsevdev->buzy = 0;

	return ret;

}

static struct file_operations touch_fops = {
	owner:		THIS_MODULE,
	open:		touch_open,
};

static struct file_operations tsusbhid_fops =
{
	owner:		THIS_MODULE,
	read:		tsusbhid_read,
	write:		tsusbhid_write,
	poll:		tsusbhid_poll,
	open:		tsusbhid_open,
	release:	tsusbhid_release,
	ioctl:		tsusbhid_ioctl,
	fasync:		tsusbhid_fasync,
};

static struct file_operations tsusbhid_evdev_fops =
{
	owner:		THIS_MODULE,
	read:		tsusbhid_evdev_read,
	write:		tsusbhid_evdev_write,
	poll:		tsusbhid_evdev_poll,
	open:		tsusbhid_evdev_open,
	release:	tsusbhid_evdev_release,
	ioctl:		tsusbhid_evdev_ioctl,
	fasync:		tsusbhid_evdev_fasync,
};

static int tsusbhid_register_chrdev(int devnum)
{
	struct file_operations *fops;
	int minor = devnum+TOUCH_MINOR_BASE;
	char *name;

	if (!(name = kmalloc(TOUCH_DEVFS_NAME_SIZE, GFP_KERNEL))) {
		TSKE("tsusbhid_register_chrdev failed! unable to alloc buffer\n");
		return -ENOMEM;
	}

	memset(name, 0, TOUCH_DEVFS_NAME_SIZE);

	spin_lock (&touch_minors_fops_lock);

	if ( devnum == TOUCH_MINORS ) {
		fops = &tsusbhid_evdev_fops;
		touch_minors_fops_table[devnum] = &tsusbhid_evdev_fops;
		snprintf( name, TOUCH_DEVFS_NAME_SIZE, TOUCH_MINOR_NAME);
	}
	else {
		fops = &tsusbhid_fops;
		touch_minors_fops_table[devnum] = &tsusbhid_fops;
		snprintf( name, TOUCH_DEVFS_NAME_SIZE, TOUCH_MINOR_NAME "%d", devnum );
	}

	spin_unlock (&touch_minors_fops_lock);

	tsusb_devfs_handle[devnum] = devfs_register(
					touch_devfs_handle,
					name,
					DEVFS_FL_DEFAULT,
					touch_devfs_major,
					minor,
					TOUCH_DEVFS_MODE,
					fops,
					NULL);

	kfree(name);

	return 0;

}

static void tsusbhid_unregister_chrdev(int devnum)
{
	spin_lock (&touch_minors_fops_lock);
	touch_minors_fops_table[devnum] = NULL;
	spin_unlock (&touch_minors_fops_lock);
	devfs_unregister(tsusb_devfs_handle[devnum]);
}

static int tsusbhid_find_devnum (struct usb_device *usbdev)
{
	int u;
	struct tsusb_hid_device *tshiddev;

	for (u = 0; u < TOUCH_MINORS && tsusb_hiddev_table[u]; u++) {

		tshiddev = tsusb_hiddev_table[u];

		if ( !tshiddev->exist  ) {
			if ( !strcmp(tshiddev->usbdev_bus_name,usbdev->bus->bus_name) &&
			     !strcmp(tshiddev->usbdev_devpath ,usbdev->devpath      ) ){
				return u;
			}
		}

	}

	for (u = 0; u < TOUCH_MINORS && tsusb_hiddev_table[u]; u++);

	if ( u == TOUCH_MINORS ) {
		TSKE("no more free tsusb devices!\n");
		return -1;
	}

	if (!(tshiddev = kmalloc(sizeof(struct tsusb_hid_device), GFP_KERNEL))) {
		TSKE("failed to alloc memory for tsusb devices!\n");
		return -1;
	}
	memset(tshiddev, 0, sizeof(struct tsusb_hid_device));

	tsusb_hiddev_table[u] = tshiddev;

	init_waitqueue_head(&tshiddev->wait);

	return u;
}

static void *tsusbhid_probe (struct usb_device *usbdev, unsigned int ifnum, const struct usb_device_id *id)
{

	struct usb_interface *intf;
	struct usb_interface_descriptor *interface;
	struct usb_endpoint_descriptor *endpoint;

	int devnum;
	struct tsusb_hid_device *tshiddev;

	int pipe;
	unsigned int maxp;

	int   buf_size = 63;
	char *buf;
	char devname[128] = { '\0', /* ... */ };

	if ( le16_to_cpu(usbdev->descriptor.idVendor) != TSUSB_VID ||
		le16_to_cpu(usbdev->descriptor.idProduct) != TSUSB_PID ) {
                TSKI("ignoring others device, vendor id 0x%x, product id 0x%x\n",
			usbdev->descriptor.idVendor,
			usbdev->descriptor.idProduct
			);
		return NULL;
	}

	if ( usbdev->descriptor.bNumConfigurations != 1 ) {
		TSKE("PROBE failed! bad NumConfigurations = %d\n",usbdev->descriptor.bNumConfigurations);
		return NULL;
	}

	intf = &usbdev->actconfig->interface[ifnum];

	if ( intf->num_altsetting != 1 ) {
		TSKE("PROBE failed! num_altsetting = %d\n",intf->num_altsetting);
		return NULL;
	}

	interface = &intf->altsetting[0];

	if (interface->bNumEndpoints != 1) {
		TSKE("PROBE failed! bad bNumEndpoints = %d\n",interface->bNumEndpoints);
		return NULL;
	}
	endpoint = interface->endpoint + 0;

	if ( !(endpoint->bEndpointAddress & 0x80) ) {
		return NULL;
	}

	if ( (endpoint->bmAttributes & 3) != 3 ) {
		return NULL;
	}

	pipe = usb_rcvintpipe(usbdev, endpoint->bEndpointAddress);

	maxp = usb_maxpacket(usbdev, pipe, usb_pipeout(pipe));

	if ( maxp > sizeof(struct tsusb_input_report) ) {
	     maxp = sizeof(struct tsusb_input_report);
	}

	devnum = tsusbhid_find_devnum(usbdev);

	if (devnum == -1) {
		return NULL;
	}

	tshiddev = tsusb_hiddev_table[devnum];

	tshiddev->input_report = kmalloc(sizeof(struct tsusb_input_report), GFP_KERNEL);
	if (!tshiddev->input_report) {
		return NULL;
	}

	tshiddev->irq = kmalloc(sizeof(struct urb), GFP_KERNEL);
	if (!tshiddev->irq) {
		kfree(tshiddev->input_report);
		return NULL;
	}

	if ( tsusbhid_register_chrdev(devnum) ) {
		kfree(tshiddev->irq);
		kfree(tshiddev->input_report);
		return NULL;
	}

	tshiddev->usbdev = usbdev;
	tshiddev->devnum = devnum;
	tshiddev->remove_pending = 0;
        tshiddev->exist = 1;

	snprintf( tshiddev->usbdev_bus_name, sizeof(tshiddev->usbdev_bus_name), usbdev->bus->bus_name);
	snprintf( tshiddev->usbdev_devpath , sizeof(tshiddev->usbdev_devpath ), usbdev->devpath      );

	usb_fill_int_urb(
		tshiddev->irq,
		usbdev,
		pipe,
		tshiddev->input_report,
		maxp,
		tsusbhid_irq,
		tshiddev,
		endpoint->bInterval
		);

        usb_submit_urb(tshiddev->irq);

	if (usbdev->descriptor.iProduct) {
		if ((buf = kmalloc(buf_size, GFP_KERNEL))) {
			memset(buf, 0, buf_size);
			if (usb_string(usbdev, usbdev->descriptor.iProduct, buf, buf_size) > 0) {
				sprintf(devname, "%s", buf);
			}
			kfree(buf);
		}
	}

	if (!strlen(devname)) sprintf(devname, DEVICE_NAME);

	sprintf(devname, "%s (%04X:%04X:%04X)",
			devname,
			usbdev->descriptor.idVendor,
			usbdev->descriptor.idProduct,
			usbdev->descriptor.bcdDevice);

	TSKI("tsusb%d -> %s on usb-%s-%s\n",
		devnum,
		devname,
		tshiddev->usbdev->bus->bus_name,
		tshiddev->usbdev->devpath
		 );

	if (tshiddev->open) WAKE_UP(&tshiddev->wait);

	tsusbhid_evdev_event(TSUSB_EVENT_PLUG,tshiddev->devnum,0,0);

	if (tsusb_evdev->open) WAKE_UP(&tsusb_evdev->wait);

	MOD_INC_USE_COUNT;

	return tshiddev;

}

static void tsusbhid_disconnect (struct usb_device *usbdev, void *ptr)
{
	struct tsusb_hid_device *tshiddev = (struct tsusb_hid_device *) ptr;

	tshiddev->remove_pending = 1;

	WAKE_UP(&tshiddev->wait);

	usb_unlink_urb(tshiddev->irq);
	kfree(tshiddev->irq);
	kfree(tshiddev->input_report);

	tsusbhid_unregister_chrdev(tshiddev->devnum);

	tshiddev->exist = 0;
	tshiddev->remove_pending = 0;

	tsusbhid_evdev_event(TSUSB_EVENT_UNPLUG,tshiddev->devnum,0,0);

	if (tsusb_evdev->open) {
		WAKE_UP(&tsusb_evdev->wait);
	}

	MOD_DEC_USE_COUNT;

}

static struct usb_device_id tsusbhid_ids [] = {
	{ USB_DEVICE(TSUSB_VID, TSUSB_PID) },
	{ }
};

MODULE_DEVICE_TABLE (usb, tsusbhid_ids);

static struct usb_driver tsusbhid_driver =
{
	name:		"tsusbhid",
	probe:		tsusbhid_probe,
	disconnect:	tsusbhid_disconnect,
	id_table:	tsusbhid_ids,
};

static int tsusbhid_evdev_init(void)
{
	int error;

	if ( !(tsusb_evdev = kmalloc(sizeof(struct tsusb_event_device), GFP_KERNEL)) ) {
		TSKE("failed to alloc memory for tsusb top devices!\n");
		return -1;
	}
	memset(tsusb_evdev, 0, sizeof(struct tsusb_event_device));

	init_waitqueue_head(&tsusb_evdev->wait);

	tsusb_evdev->devnum = TOUCH_MINORS;

	error = tsusbhid_register_chrdev(tsusb_evdev->devnum);
	if (error) {
		TSKE("Not able to register char device for top device.\n");
		return error;
	}

	return 0;

}

static void tsusbhid_evdev_exit(void)
{

	tsusbhid_unregister_chrdev(tsusb_evdev->devnum);

	kfree(tsusb_evdev);

}

static int __init tsusbhid_init (void)
{
	int error;

	error = devfs_register_chrdev(touch_devfs_major, TOUCH_MAJOR_NAME, &touch_fops);
	if (error) {
		TSKE("unable to register major %d for touch devices, error %d\n", touch_devfs_major, error);
		return -EBUSY;
	}

	touch_devfs_handle = devfs_mk_dir(NULL, TOUCH_MAJOR_NAME, NULL);

	tsusbhid_evdev_init();

	error = usb_register(&tsusbhid_driver);
	if (error) {
		TSKE("unable to register usb driver, error %d\n", error );
		return error;
	}

	TSKI(DRIVER_VERSION ":" DRIVER_DESC "\n");

	return 0;
}

static void __exit tsusbhid_exit (void)
{
	tsusbhid_evdev_exit();
        devfs_unregister(touch_devfs_handle);

	if (devfs_unregister_chrdev(touch_devfs_major, TOUCH_MAJOR_NAME)) {
		TSKE("can't unregister char major %d\n", touch_devfs_major);
	}

	usb_deregister (&tsusbhid_driver);

}

MODULE_AUTHOR		( DRIVER_AUTHOR );
MODULE_DESCRIPTION	( DRIVER_DESC );
MODULE_LICENSE		( DRIVER_LICENSE);

module_init (tsusbhid_init);
module_exit (tsusbhid_exit);
