#ifndef __TSUSBHID_H
#define __TSUSBHID_H

#define TSUSB_DRIVER_VERCODE	0x010002
#define TSUSB_DRIVER_VERSION	"v1.0.2"

#define TSUSB_MAJOR_NAME	"touch"
#define TSUSB_MINOR_NAME	"tsusb"
#define TSUSB_MINOR_BASE	0
#define TSUSB_MINORS		255

#define TSUSB_VID		0x16FD
#define TSUSB_PID		0x5453

#define TSUSB_FEATURE_ID_INIT	0x7F
#define TSUSB_FEATURE_ID_SAVE	0x00

#define TSUSB_LIST_BUFFER_SIZE	64

#define TSUSB_EVENT_PLUG	0xDE
#define TSUSB_EVENT_UNPLUG	0xDD
#define TSUSB_EVENT_INPUT_DATA	0x00
#define TSUSB_EVENT_ENABLE	0xCE
#define TSUSB_EVENT_DISABLE	0xCD

struct tsusb_event_packet{
	unsigned short int	id;
	unsigned short int	v1;
	unsigned short int	v2;
	unsigned short int	v3;
} __attribute__((packed));

struct tsusb_input_report{
	unsigned char		state;
	unsigned short int	xdata;
	unsigned short int	ydata;
} __attribute__((packed));

struct tsusb_feature_report {
	unsigned char		feature_id;
	unsigned short int	calibration_x1;
	unsigned short int	calibration_x2;
	unsigned short int	calibration_x3;
	unsigned short int	calibration_y1;
	unsigned short int	calibration_y2;
	unsigned short int	calibration_y3;
	unsigned char		panel_direction;
	unsigned char		sleep_time;
	unsigned char		virtual_button;
	unsigned char		right_button_function;
	unsigned char		right_button_auto_mode;
	unsigned char		right_button_auto_stime;
	unsigned char		right_button_auto_etime;
	unsigned char		right_button_auto_beep;
	unsigned char		right_button_auto_range;
	unsigned char		feedback_sound;
	unsigned short int	feedback_tone;
	unsigned short int	feedback_duration;
	unsigned short int	raw_x;
	unsigned short int	raw_y;
	unsigned char		raw_touch;
	unsigned char		raw_count;
	unsigned char		error_code1;
	unsigned char		error_code2;
	unsigned char		error_code3;
	unsigned char		error_code4;
} __attribute__((packed));

#define TSUSBIOC_TYPE		'U'
#define TSUSBIOC_GET_VERSION	_IOR (TSUSBIOC_TYPE, 0x01, int)
#define TSUSBIOC_MAX_NR		                     0x01

typedef short tshidioc_id_type;

#define TSHIDIOC_TYPE		'H'
#define TSHIDIOC_GET_VERSION	_IOR (TSHIDIOC_TYPE, 0x01, int)
#define TSHIDIOC_GET_ID		_IOR (TSHIDIOC_TYPE, 0x02, tshidioc_id_type[TSHIDIOC_ID_SIZE])
#define TSHIDIOC_GET_FEATURE	_IOR (TSHIDIOC_TYPE, 0x03, struct tsusb_feature_report)
#define TSHIDIOC_SET_FEATURE	_IOW (TSHIDIOC_TYPE, 0x04, struct tsusb_feature_report)
#define TSHIDIOC_SET_ENABLE	_IO  (TSHIDIOC_TYPE, 0x05)
#define TSHIDIOC_SET_DISABLE	_IO  (TSHIDIOC_TYPE, 0x06)
#define TSHIDIOC_MAX_NR		                     0x06

#define TSHIDIOC_ID_VENDOR	0
#define TSHIDIOC_ID_PRODUCT	1
#define TSHIDIOC_ID_REVISION	2
#define TSHIDIOC_ID_BUSNUM	3
#define TSHIDIOC_ID_DEVNUM	4
#define TSHIDIOC_ID_SIZE	5

#define TSHIDIOC_INPUT_REPORT	0
#define TSHIDIOC_OUTPUT_REPORT	1
#define TSHIDIOC_FEATURE_REPORT	2

#ifdef __KERNEL__

//*********************************************
// NOTE: DO NOT CHANGE THE FOLLOWINGS SETTINGS!
//*********************************************
#define TSUSB_CLASS_NAME         "TsUsbHid"
#define TSUSB_CLASS_DEVICE_NAME  "TsUsb"
//*********************************************

#define	TSHID_REQ_GET		USB_DIR_IN  + USB_TYPE_CLASS + USB_RECIP_INTERFACE
#define	TSHID_REQ_SET		USB_DIR_OUT + USB_TYPE_CLASS + USB_RECIP_INTERFACE

#define	TSHID_REQ_GET_REPORT	0x01
#define	TSHID_REQ_GET_IDEL	0x02
#define	TSHID_REQ_GET_PROTOCOL	0x03

#define	TSHID_REQ_SET_REPORT	0x09
#define	TSHID_REQ_SET_IDEL	0x0A
#define	TSHID_REQ_SET_PROTOCOL	0x0B

#define	TSHID_REPORT_TYPE_INPUT		0x01
#define	TSHID_REPORT_TYPE_OUTPUT	0x02
#define	TSHID_REPORT_TYPE_FEATURE	0x03

#define	TSHID_REQ_TYPE_ID(type,id)	(type<<8|id)

struct tsusb_hid_device_list {
	struct tsusb_input_report buffer[TSUSB_LIST_BUFFER_SIZE];
	int head;
	int tail;
	struct fasync_struct *fasync;
	struct tsusb_hid_device *tshiddev;
	struct tsusb_hid_device_list *next;
};

struct tsusb_hid_device {
	struct usb_device *usbdev;
	wait_queue_head_t wait;
	int devnum;
	int exist;
	int open;
	int buzy;
	int remove_pending;
	struct tsusb_feature_report feature;

	struct urb *irq;
	struct tsusb_input_report *input_report;
	dma_addr_t                 input_report_dma;

	struct tsusb_hid_device_list *list;

	char usbdev_bus_name[20];
	char usbdev_devpath [20];

};

struct tsusb_event_device_list {
	struct tsusb_event_packet buffer[TSUSB_LIST_BUFFER_SIZE];
	int head;
	int tail;
	struct tsusb_event_device *tsevdev;
	struct fasync_struct *fasync;
	struct tsusb_event_device_list *next;
};

struct tsusb_event_device {
	wait_queue_head_t wait;
	int devnum;
	int open;
	int buzy;
	struct tsusb_event_device_list *list;
};

#endif /* __KERNEL__ */

#endif /* __TSUSBHID_H */
