/*
* tsusb_drv - TSUSB Touchscreen X Input Driver
*
* Copyright 2007 by CometrueWorks
*
* Permission to use, copy, modify, distribute, and sell this software and its
* documentation for any purpose is  hereby granted without fee, provided that
* the  above copyright   notice appear  in   all  copies and  that both  that
* copyright  notice   and   this  permission   notice  appear  in  supporting
* documentation, and that   the  name of  Frederic   Lepied not  be  used  in
* advertising or publicity pertaining to distribution of the software without
* specific,  written      prior  permission.     Frederic  Lepied   makes  no
* representations about the suitability of this software for any purpose.  It
* is provided "as is" without express or implied warranty.
*
* FREDERIC  LEPIED DISCLAIMS ALL   WARRANTIES WITH REGARD  TO  THIS SOFTWARE,
* INCLUDING ALL IMPLIED   WARRANTIES OF MERCHANTABILITY  AND   FITNESS, IN NO
* EVENT  SHALL FREDERIC  LEPIED BE   LIABLE   FOR ANY  SPECIAL, INDIRECT   OR
* CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE,
* DATA  OR PROFITS, WHETHER  IN  AN ACTION OF  CONTRACT,  NEGLIGENCE OR OTHER
* TORTIOUS  ACTION, ARISING    OUT OF OR   IN  CONNECTION  WITH THE USE    OR
* PERFORMANCE OF THIS SOFTWARE.
*
*/
/* $XFree86: xc/programs/Xserver/hw/xfree86/input/tsusb/tsusb.c,v 1.0.1 tsi Exp $ */

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#define _tsusb_C_

#include <asm/types.h>
#include <linux/input.h>


/*****************************************************************************
 *	Standard Headers
 ****************************************************************************/
#include <misc.h>
#include <xf86.h>
#ifndef	NEED_XF86_TYPES
#define	NEED_XF86_TYPES
#endif

/*** DO NOT INCLUSE xf86_ansic.h from XOrg 7.1, insted include errno.h, fcntl.h ***/
//#include <xf86_ansic.h>
//#include <strings.h>
//#include <stdio.h>
#include <errno.h>
#include <fcntl.h>
/*** DO NOT INCLUSE xf86_ansic.h from XOrg 7.1, insted include errno.h, fcntl.h ***/

#include <xf86_OSproc.h>
#include <xf86Xinput.h>
#include <xisb.h>
#include <exevents.h>

#include "xf86Version.h"

#if	XF86_VERSION_CURRENT < XF86_VERSION_NUMERIC(4,0,0,0,0)
	#error "*** XFREE86 VERSION 4.0 OR LATER REQUIRED!"
#endif

#ifndef	XFree86LOADER
	#error "*** UNDEFINED: XFree86LOADER"
#endif

#include "xf86Module.h"


/*****************************************************************************
 *	Local Headers
 ****************************************************************************/
#include "tsusbhid.h"
#include "tsusb.h"
#include "tsini.h"

#define	TSDEBUG				0
#define	TSDEBUG_Convert			0
#define	TSDEBUG_LoadConfiguration	0
#define	TSDEBUG_TsUSB_SendTouchEvent	0
#define	TSDEBUG_TsUSB_ReadInput		0


#define	TSEQU_DEVICE	"/dev/" TSUSB_MAJOR_NAME "/" TSUSB_MINOR_NAME

#define TS_ADCIN_MIN		0
#define TS_ADCIN_MAX		1023

#define E_VBT_LIFT_OFF		55
#define E_VBT_TOUCH_DOWN	55

#define MOUSE_BUTTON_LEFT	1
#define MOUSE_BUTTON_RIGHT	2
#define MOUSE_BUTTON_MIDDLE	4

#define XMICE_BUTTON_LEFT	1
#define XMICE_BUTTON_MIDDLE	2
#define XMICE_BUTTON_RIGHT	3
#define XMICE_BUTTON_TOTAL	3

#undef	SYSCALL
#undef	read
#undef	write
#undef	close
#define	SYSCALL(call)		call
#define	read(fd, ptr, num) 	xf86ReadSerial(fd, ptr, num)
#define	write(fd, ptr, num)	xf86WriteSerial(fd, ptr, num)
#define	close(fd)		xf86CloseSerial(fd)

static const char *default_options[] = { NULL };

#define BEEP_AUTOLEFTBUTTON	TsUSB_Beep_AUTOLEFTBUTTON(priv)
#define BEEP_AUTORIGHTBUTTON	TsUSB_Beep_AUTORIGHTBUTTON(priv)

static CARD32
TsUSB_Beep_TimeoutProc(
	OsTimerPtr timer,
	CARD32     now,
	pointer    arg
	)
{

	CARD32 millisecond = 0;

	TsPrivatePtr priv = (TsPrivatePtr)(arg);

	if ( priv->BellTimerCounter < E_BELL_RECORD_MAX ) {
		xf86SoundKbdBell(
			priv->BellVolume[priv->BellTimerCounter],
			priv->BellPitch[priv->BellTimerCounter],
			priv->BellDuration[priv->BellTimerCounter]
			);
		millisecond = 100;
		priv->BellTimerCounter++;
	}
	return (millisecond);
}

static void
TsUSB_Beep_TimerStart( TsPrivatePtr priv )
{
	priv->BellTimerPointer = TimerSet(priv->BellTimerPointer,
						0,
						1,
						TsUSB_Beep_TimeoutProc,
						priv
						);
}

static void
TsUSB_Beep_AUTOLEFTBUTTON( TsPrivatePtr priv )
{
	priv->BellVolume[0]=100; priv->BellPitch[0]=1600; priv->BellDuration[0]=50;
	priv->BellVolume[1]=100; priv->BellPitch[1]=1300; priv->BellDuration[1]=50;
	priv->BellVolume[2]=100; priv->BellPitch[2]=1000; priv->BellDuration[2]=50;
	priv->BellTimerCounter =0;
	TsUSB_Beep_TimerStart(priv);
}

static void
TsUSB_Beep_AUTORIGHTBUTTON( TsPrivatePtr priv )
{
	priv->BellVolume[0]=100; priv->BellPitch[0]=1000; priv->BellDuration[0]=50;
	priv->BellVolume[1]=100; priv->BellPitch[1]=1300; priv->BellDuration[1]=50;
	priv->BellVolume[2]=100; priv->BellPitch[2]=1600; priv->BellDuration[2]=50;
	priv->BellTimerCounter = 0;
	TsUSB_Beep_TimerStart(priv);
}

static void
TsUSB_Beep_TOUCH( TsPrivatePtr priv )
{
	xf86SoundKbdBell(
		priv->use_TouchBeepVolume,
		priv->use_TouchBeepTone,
		priv->use_TouchBeepDuration
		);
}

static CARD32
TsUSB_VBTimeoutProc(
	OsTimerPtr timer,
	CARD32     now,
	pointer    arg
	)
{
	TsPrivatePtr	priv  = (TsPrivatePtr)(arg);

	if ( priv->VBTimerMode == 2 ){
		priv->db_AutoRightButtonGo = 1;
		priv->VBTimerMode = 0;
		return 0;
	}

	if ( priv->VBTimerMode == 3 ){
		priv->db_AutoLeftButtonGo = 0;
		priv->VBTimerMode = 0;
		if (priv->use_RightButtonAutoBeep==1) TsUSB_Beep_AUTOLEFTBUTTON(priv);
		return 0;
	}

	if (priv->VBTimerMode==1) priv->dw_RBF_Action = E_RBF_ACTION_OFF;

	priv->VBTimerMode = 0;

	return 0;
}

static void
TsUSB_SendTouchEvent( int mode, int new_state, int new_x, int new_y, LocalDevicePtr local, TsPrivatePtr priv)
{

	int	zSound, zButton, buttonsDelta;

	if ( priv->use_TouchBeepMode != TS_FS_NO_SOUND ) {
		switch (priv->use_TouchBeepMode)
		{
			case TS_FS_TOUCH_ONLY:
				zSound = ( new_state!=0 && priv->raw_state==0 );
				break;
			case TS_FS_LIFT_ONLY:
				zSound = ( new_state==0 && priv->raw_state!=0 );
				break;
			case TS_FS_TOUCH_AND_LIFT:
				zSound = ( new_state!=0 && priv->raw_state==0 )||
					 ( new_state==0 && priv->raw_state!=0 );
				break;
			default:
				zSound = FALSE;
				break;
		}
		if (zSound) TsUSB_Beep_TOUCH(priv);
	}

	if ( priv->use_RightButtonFunction == TS_RB_FUNCTION_ON ) {
		zButton = (priv->dw_RBF_Action==E_RBF_ACTION_OFF)? MOUSE_BUTTON_LEFT: MOUSE_BUTTON_RIGHT;
	}
	else{
		zButton = MOUSE_BUTTON_LEFT;
	}

	priv->raw_buttons = 0;

	switch (priv->use_VirtualButton)
	{
		case TS_VB_STREAM:
			priv->dw_RBF_Action = E_RBF_ACTION_OFF;
			break;
		case TS_VB_TOUCH_DOWN:
			if ( new_state!=0 && priv->raw_state==0 ) {
				priv->raw_buttons |= zButton;
				if ( zButton == MOUSE_BUTTON_RIGHT ) priv->dw_RBF_Action = E_RBF_ACTION_OFF;
				priv->VBTimerMode = 255;
				priv->VBTimerPointer = TimerSet(priv->VBTimerPointer,
								0,
								E_VBT_TOUCH_DOWN,
								TsUSB_VBTimeoutProc,
								priv
								);
			}
			else {
				if (priv->VBTimerMode == 255){
					new_x = priv->raw_x;
                			new_y = priv->raw_y;
					priv->raw_buttons = priv->now_buttons;
				}
			}
			break;
		case TS_VB_LIFT_OFF:
			if ( new_state==0 && priv->raw_state!=0 ) {
				priv->raw_buttons |= zButton;
				priv->VBTimerMode = 1;
				priv->VBTimerPointer = TimerSet(priv->VBTimerPointer,
								0,
								E_VBT_LIFT_OFF,
								TsUSB_VBTimeoutProc,
								priv
								);
			}
			break;
		case TS_VB_DRAG_DROP:
			if (	priv->use_RightButtonFunction == TS_RB_FUNCTION_ON &&
				priv->use_RightButtonAutoMode == TS_RB_MODE_AUTOMETIC &&
				priv->dw_RBF_Action           == E_RBF_ACTION_OFF ) {
				if ( new_state!=0 && priv->raw_state==0 ) {
					if ( priv->VBTimerMode == 3 ){
						TimerCancel(priv->VBTimerPointer);
						priv->VBTimerMode = 0;
						zButton = MOUSE_BUTTON_RIGHT;
						priv->db_AutoRightButtonGo = 0;
						priv->db_AutoLeftButtonGo  = 2;
						priv->dw_RBF_Action = E_RBF_ACTION_ON;
					}
				else{
					priv->dw_AutoRightButtonX1 = new_x - priv->use_RightButtonAutoRange;
					priv->dw_AutoRightButtonY1 = new_y - priv->use_RightButtonAutoRange;
					priv->dw_AutoRightButtonX2 = new_x + priv->use_RightButtonAutoRange;
					priv->dw_AutoRightButtonY2 = new_y + priv->use_RightButtonAutoRange;
					priv->db_AutoRightButtonGo = 0;
					priv->db_AutoLeftButtonGo  = 0;
					priv->VBTimerMode = 2;
					priv->VBTimerPointer = TimerSet(priv->VBTimerPointer,
									0,
									(CARD32)((priv->use_RightButtonAutoSTime+1)*500),
									TsUSB_VBTimeoutProc,
									priv
									);
				}
			}
			else{
				switch  (priv->db_AutoRightButtonGo)
				{
					case 0:
						break;
					case 1:
						if (new_x >= priv->dw_AutoRightButtonX1 &&
							new_y >= priv->dw_AutoRightButtonY1 &&
							new_x <= priv->dw_AutoRightButtonX2 &&
							new_y <= priv->dw_AutoRightButtonY2){
							zButton = MOUSE_BUTTON_LEFT|MOUSE_BUTTON_RIGHT;
							priv->db_AutoRightButtonGo = 2;
						}
						else{
							priv->db_AutoRightButtonGo = 0;
						}
						break;
					case 2:
						zButton = MOUSE_BUTTON_RIGHT;
						priv->db_AutoRightButtonGo = 0;
						priv->db_AutoLeftButtonGo  = 1;
						priv->dw_RBF_Action = E_RBF_ACTION_ON;
						if (priv->use_RightButtonAutoBeep==1) TsUSB_Beep_AUTORIGHTBUTTON(priv);
						break;
				}
			}
		}

		if  ( new_state != 0 ){
			priv->raw_buttons |= zButton;
		}
		else{
			priv->dw_RBF_Action = E_RBF_ACTION_OFF;
			if ( priv->VBTimerMode == 2 ){
				TimerCancel(priv->VBTimerPointer);
				priv->VBTimerMode = 0;
			}
			else{
				if ( priv->db_AutoLeftButtonGo == 1 ){
					priv->db_AutoLeftButtonGo = 0;
					priv->VBTimerMode = 3;
					priv->VBTimerPointer =
						TimerSet(priv->VBTimerPointer,
							0,
							(priv->use_RightButtonAutoETime+1)*500,
							TsUSB_VBTimeoutProc,
							priv
							);
				}
				else{
					if ( priv->db_AutoLeftButtonGo == 2 ){
						priv->db_AutoLeftButtonGo = 0;
						if (priv->use_RightButtonAutoBeep==1) TsUSB_Beep_AUTOLEFTBUTTON(priv);
					}
				}
			}
		}
		break;
	}

	priv->raw_x	= new_x;
	priv->raw_y	= new_y;
	priv->raw_state = new_state;

	xf86PostMotionEvent(local->dev, TRUE, 0, 5, mode, new_state, new_x, new_y, priv->devnum);

	#define POST_BUTTON(n,f)    xf86PostButtonEvent(local->dev, TRUE, n, f, 0, 2, 21, 22)
	#define POST_BUTTON_UP(n)   POST_BUTTON(n,FALSE)
	#define POST_BUTTON_DOWN(n) POST_BUTTON(n,TRUE )

	buttonsDelta = priv->now_buttons ^ priv->raw_buttons;
	if (buttonsDelta) {
		if (buttonsDelta & MOUSE_BUTTON_LEFT) {
			if (priv->raw_buttons & MOUSE_BUTTON_LEFT) {
				POST_BUTTON_DOWN(XMICE_BUTTON_LEFT);
			}
			else {
				POST_BUTTON_UP  (XMICE_BUTTON_LEFT);
			}
		}
		if (buttonsDelta & MOUSE_BUTTON_MIDDLE) {
			if (priv->raw_buttons & MOUSE_BUTTON_MIDDLE) {
				POST_BUTTON_DOWN(XMICE_BUTTON_MIDDLE);
			}
			else {
				POST_BUTTON_UP  (XMICE_BUTTON_MIDDLE);
			}
		}
		if (buttonsDelta & MOUSE_BUTTON_RIGHT) {
			if (priv->raw_buttons & MOUSE_BUTTON_RIGHT) {
				POST_BUTTON_DOWN(XMICE_BUTTON_RIGHT);
			}
			else {
				POST_BUTTON_UP  (XMICE_BUTTON_RIGHT);
			}
		}
		priv->now_buttons = priv->raw_buttons;
	}

}

static Bool
TsUSB_Convert(
	LocalDevicePtr	local,
	int	first,
	int	num,
	int	v0,
	int	v1,
	int	v2,
	int	v3,
	int	v4,
	int	v5,
	int	*x,
	int	*y)
{
	TsDevicePtr tsdev = (TsDevicePtr) local->private;
	TsPrivatePtr priv;
	int input_x, input_y;

	if (first != 0 || num != 5) {
		return FALSE;
	}

	input_x = v2;
	input_y = v3;

	priv = (TsPrivatePtr) tsdev->priv[v4];

	*x = input_x * priv->screen_width  / TS_ADCIN_MAX;
	*y = input_y * priv->screen_height / TS_ADCIN_MAX;

	xf86XInputSetScreen(local, priv->screen_no, *x, *y);

	return TRUE;

}

static Bool TsUSB_GetFeatureReport(TsPrivatePtr priv)
{
	int err;
	int fd;

	SYSCALL(fd = open(priv->input_dev, O_RDWR, 0));


	if (fd < 0)
	{
		xf86Msg(X_ERROR,"TSUSB: Unable to open device '%s'\n",priv->input_dev);
		return FALSE;
	}

	SYSCALL(err = ioctl(fd, TSHIDIOC_GET_FEATURE, &priv->feature));
	SYSCALL(close(fd));

	if (err) {
		xf86Msg(X_ERROR,"TSUSB: Failed on get feature report from device '%s'\n",priv->input_dev);
		return FALSE;
	}

	return TRUE;
}

static Bool TsUSB_AskFeatures(TsPrivatePtr priv)
{
	int ok;

        ok = TsUSB_GetFeatureReport(priv);

	if (!ok) {
		xf86Msg(X_ERROR,"TSUSB: Unable to get feature report from device '%s', using default.\n",priv->input_dev);
		priv->use_TouchBeepMode		= TSINI_TOUCH_BEEP_MODE;
		priv->use_TouchBeepVolume	= TSINI_TOUCH_BEEP_VOLUME;
		priv->use_TouchBeepTone		= TSINI_TOUCH_BEEP_TONE;
		priv->use_TouchBeepDuration	= TSINI_TOUCH_BEEP_DURATION;
		priv->use_VirtualButton		= TSINI_VIRTUAL_BUTTON;
		priv->use_RightButtonFunction	= TSINI_RIGHT_BUTTON_FUNCTION;
		priv->use_RightButtonAutoMode	= TSINI_RIGHT_BUTTON_AUTO_MODE;
		priv->use_RightButtonAutoSTime	= TSINI_RIGHT_BUTTON_AUTO_STIME;
		priv->use_RightButtonAutoETime	= TSINI_RIGHT_BUTTON_AUTO_ETIME;
		priv->use_RightButtonAutoBeep	= TSINI_RIGHT_BUTTON_AUTO_BEEP;
		priv->use_RightButtonAutoRange	= TSINI_RIGHT_BUTTON_AUTO_RANGE;
		return FALSE;
	}

	priv->use_TouchBeepMode		= priv->feature.feedback_sound;
	priv->use_TouchBeepVolume	= TSINI_TOUCH_BEEP_VOLUME;
	priv->use_TouchBeepTone		= priv->feature.feedback_tone;
	priv->use_TouchBeepDuration	= priv->feature.feedback_duration;
	priv->use_VirtualButton		= priv->feature.virtual_button;
	priv->use_RightButtonFunction	= priv->feature.right_button_function;
	priv->use_RightButtonAutoMode	= priv->feature.right_button_auto_mode;
	priv->use_RightButtonAutoSTime	= priv->feature.right_button_auto_stime;
	priv->use_RightButtonAutoETime	= priv->feature.right_button_auto_etime;
	priv->use_RightButtonAutoBeep	= priv->feature.right_button_auto_beep;
	priv->use_RightButtonAutoRange	= priv->feature.right_button_auto_range;

	return TRUE;
}

static TsPrivatePtr
TsUSB_Allocate_Private(int devnum)
{
	TsPrivatePtr	priv;

	priv = xalloc(sizeof(TsPrivateRec));
	if (!priv) {
		return NULL;
	}

	priv->input_dev = xalloc(strlen(TSEQU_DEVICE)+4);
	if (!priv->input_dev) {
		xfree(priv);
		return NULL;
	}

	snprintf( priv->input_dev, strlen(TSEQU_DEVICE)+4, TSEQU_DEVICE "%d" , devnum );

	priv->fd = -1;
	priv->devnum = devnum;
	priv->inited = 0;
	priv->screen_no = 0;
	priv->screen_width = screenInfo.screens[priv->screen_no]->width;
	priv->screen_height = screenInfo.screens[priv->screen_no]->height;
	priv->raw_x = 0;
	priv->raw_y = 0;
	priv->raw_state = 0;
	priv->raw_buttons = 0;
	priv->now_buttons = 0;
	priv->device_enabled = 1;
	priv->dw_RBF_Action = E_RBF_ACTION_OFF;
	priv->db_AutoRightButtonGo = 0;
	priv->db_AutoLeftButtonGo = 0;
	priv->dw_AutoRightButtonX1 = 0;
	priv->dw_AutoRightButtonY1 = 0;
	priv->dw_AutoRightButtonX2 = 0;
	priv->dw_AutoRightButtonY2 = 0;
	priv->VBTimerMode = 0;
	priv->VBTimerPointer = NULL;
	priv->BellTimerCounter = 0;
	priv->BellTimerPointer = NULL;

	return priv;
}

void TsUSB_ReadInput(LocalDevicePtr local)
{

	TsDevicePtr	tsdev = (TsDevicePtr)(local->private);

	struct tsusb_event_packet *event;

	int packetLength = sizeof(struct tsusb_event_packet);

	int len, pos, remaining;

	int devnum;

	remaining = sizeof(tsdev->buffer) - tsdev->bufpos;

	SYSCALL(len = read(local->fd, tsdev->buffer + tsdev->bufpos, remaining));

	if (len <= 0)
	{
		ErrorF("Error reading TSUSB device : %s\n", strerror(errno));
		return;
	}

	tsdev->bufpos += len;

	pos = 0;

	while ((tsdev->bufpos - pos) >=  packetLength)
	{
		event = (struct tsusb_event_packet *)(tsdev->buffer + pos);

		switch (event->id) {
			case	TSUSB_EVENT_PLUG:
				devnum = event->v1;
				tsdev->priv[devnum] = TsUSB_Allocate_Private(devnum);
				if (!tsdev->priv[devnum]) {
					xf86Msg(X_ERROR,"TSUSB: Unable to allocate private data for devnum %d\n",devnum);
				}
				else {
					TsUSB_AskFeatures(tsdev->priv[devnum]);
				}
				break;
			case	TSUSB_EVENT_UNPLUG:
				devnum = event->v1;
				if (tsdev->priv[devnum]) {
					if (tsdev->priv[devnum]->input_dev)  {
						xfree(tsdev->priv[devnum]->input_dev);
						tsdev->priv[devnum]->input_dev = NULL;
					}
					xfree(tsdev->priv[devnum]);
					tsdev->priv[devnum] = NULL;
				}
				break;
			case	TSUSB_EVENT_INPUT_DATA:
				devnum = (event->v1 >> 8 & 0xff);
				if (tsdev->priv[devnum]) {
					if ( tsdev->priv[devnum]->device_enabled ) {
						TsUSB_SendTouchEvent(
							Absolute,
							(event->v1&0xff),
							event->v2,
							event->v3,
							local,
							tsdev->priv[devnum]
							);
					}
				}
				break;
			case	TSUSB_EVENT_ENABLE:
				devnum = (event->v1 >> 8 & 0xff);
				if (tsdev->priv[devnum]) {
					tsdev->priv[devnum]->device_enabled = 1;
					TsUSB_AskFeatures(tsdev->priv[devnum]);
				}
				break;
			case	TSUSB_EVENT_DISABLE:
				devnum = (event->v1 >> 8 & 0xff);
				if (tsdev->priv[devnum]) {
					tsdev->priv[devnum]->device_enabled = 0;
				}
				break;
			default:
				break;
		}
		pos += packetLength;
	}

	if (pos)
	{
		if (pos < tsdev->bufpos)
		{
			memmove( tsdev->buffer, tsdev->buffer+pos, tsdev->bufpos-pos );
			tsdev->bufpos -= pos;
		}
		else
		{
			tsdev->bufpos = 0;
		}
	}

}

static void
TsUSB_PtrFeedbackProc(DeviceIntPtr device, PtrCtrl *ctrl)
{
}

static Bool TsUSB_Detect(int fd)
{

	int version;
	int err;

	SYSCALL(err = ioctl(fd, TSUSBIOC_GET_VERSION, &version));

	if (err) {
		xf86Msg(X_ERROR,"TSUSB: Failed on get Kernel Input Driver Version (error=%d)\n",err);
		return FALSE;
	}

	xf86Msg(X_INFO,"TSUSB: Kernel Input Driver Version is %d.%d.%d\n",
			version >> 16,	(version >> 8) & 0xff,	version & 0xff);
	return TRUE;

}

static Bool
TsUSB_DeviceControl(
	DeviceIntPtr	dev,
	int		mode)
{

	CARD8	map[XMICE_BUTTON_TOTAL+1];
	int	nValuators;
	int	nButtons;
	int	loop;

	LocalDevicePtr	local = (LocalDevicePtr) dev->public.devicePrivate;
	TsDevicePtr	tsdev  = (TsDevicePtr)(local->private);

	switch(mode) {

		case DEVICE_INIT:
		{
			nValuators = 6;
			nButtons  = XMICE_BUTTON_TOTAL;

			for(loop=1; loop<=nButtons; loop++) map[loop] = loop;

			if (InitButtonClassDeviceStruct(dev, nButtons, map) == FALSE) {
				xf86Msg(X_ERROR,"TSUSB: Unable to allocate ButtonClassDeviceStruct\n");
				return !Success;
			}

			if (InitFocusClassDeviceStruct(dev) == FALSE) {
				xf86Msg(X_ERROR,"TSUSB: Unable to allocate FocusClassDeviceStruct\n");
				return !Success;
			}

			if (InitPtrFeedbackClassDeviceStruct(dev,TsUSB_PtrFeedbackProc) == FALSE) {
				xf86Msg(X_ERROR,"TSUSB: Unable to initialize Feedback Class Device\n");
				return !Success;
			}

			if (InitValuatorClassDeviceStruct(
				dev,
				nValuators,
				xf86GetMotionEvents,
				local->history_size,
				Absolute
				) == FALSE) {
				xf86Msg(X_ERROR,"TSUSB: Unable to allocate ValuatorClassDeviceStruct\n");
				return !Success;
			}

			for(loop=0; loop<=nValuators; loop++){
				InitValuatorAxisStruct(
					dev,
					loop,
					0,
					0xFFFFFFFF,
					0x7FFFFFFF,
					0,
					0xFFFFFFFF
					);
			}

			if (InitProximityClassDeviceStruct(dev) == FALSE) {
				xf86Msg(X_ERROR,"TSUSB: Unable to initialize Proximity Class Device\n");
				return !Success;
			}

			xf86MotionHistoryAllocate(local);

			return Success;
		}

		case DEVICE_ON:
		{
			if (local->fd < 0)
			{
				xf86Msg(X_INFO,"TSUSB: Detecting Touchscreen Controller...\n");

				SYSCALL(local->fd = open(tsdev->devfs_name, O_RDWR, 0));

				if (local->fd < 0)
				{
					xf86Msg(X_ERROR,"TSUSB: Unable to open device '%s'\n",tsdev->devfs_name);
					return !Success;
				}

				if ( !TsUSB_Detect(local->fd) )
				{
					xf86Msg(X_ERROR,"TSUSB: Can not find Touchscreen Controller on %s!\n", tsdev->devfs_name);
					SYSCALL(close(local->fd));
					return !Success;
				}

				xf86Msg(X_INFO,"TSUSB: Touchscreen Controller Found.\n");

				xf86AddEnabledDevice(local);

				dev->public.on = TRUE;

			}
			return Success;
		}

		case DEVICE_OFF:
		{
			if (local->fd != -1)
			{
				ErrorF("            TsUSB_DeviceControl: xf86RemoveEnabledDevice\n");
				xf86RemoveEnabledDevice(local);
				SYSCALL(close(local->fd));
				local->fd = -1;
			}
			dev->public.on = FALSE;
			return Success;
		}

		case DEVICE_CLOSE:
		{
			if (local->fd != -1) {
				xf86RemoveEnabledDevice(local);
				SYSCALL(close(local->fd));
				local->fd = -1;
			}
			dev->public.on = FALSE;
			return Success;
		}

		default:
		{
			xf86Msg(X_ERROR,"TSUSB: Unsupported Control Mode %d\n", mode);
			return !Success;
		}
	}
}

static LocalDevicePtr
TsUSB_Allocate(
	InputDriverPtr	drv,
	IDevPtr		dev)
{
	LocalDevicePtr	local;
	TsDevicePtr	tsdev;
	int i;

	tsdev = xalloc(sizeof(TsDeviceRec));
	if (!tsdev) {
		return NULL;
	}

	local = xf86AllocateInput(drv, 0);

	if (!local) {
		xfree(tsdev);
		return NULL;
	}

	tsdev->devfs_name = strdup(TSEQU_DEVICE);

	tsdev->bufpos = 0;
	for ( i=0; i<TSUSB_MINORS; i++ ) {
		tsdev->priv[i] = NULL;
	}

	local->name = dev->identifier;
	local->type_name = XI_TOUCHSCREEN;
	local->flags = 0;
	local->device_control = TsUSB_DeviceControl;
	local->read_input = TsUSB_ReadInput;
	local->control_proc = NULL;
	local->close_proc = NULL;
	local->switch_mode = NULL;
	local->conversion_proc = TsUSB_Convert;
	local->reverse_conversion_proc = NULL;
	local->fd = -1;
	local->atom = 0;
	local->dev = NULL;
	local->private = tsdev;
	local->history_size = 0;

	return local;
}

static InputInfoPtr
TsUSB_Init(
	InputDriverPtr	drv,
	IDevPtr		dev,
	int		flags)
{
	LocalDevicePtr	local = NULL;
	TsDevicePtr	tsdev  = NULL;

	char *str;

	xf86Msg(X_INFO, "TSUSB: Initializing ...\n");

	local = TsUSB_Allocate(drv,dev);
	if (!local) {
		return NULL;
	}

	tsdev = local->private;

	local->conf_idev = dev;

	xf86CollectInputOptions(local, default_options, NULL);

	xf86ProcessCommonOptions(local, local->options);

	str = xf86FindOptionValue(local->options, "Device");
	if (!str) {
		xf86Msg(X_ERROR, "%s: No Device specified in module config.\n",dev->identifier);
		if (tsdev) {
			if (tsdev->devfs_name) {
				xfree(tsdev->devfs_name);
			}
			xfree(tsdev);
		}
		return local;
	}
	tsdev->devfs_name = strdup(str);

	(void)xf86LoadKernelModule("tsusbhid");

	local->flags |= XI86_POINTER_CAPABLE | XI86_CONFIGURED;

	return local;

}

static void
TsUSB_Uninit(
	InputDriverPtr	drv,
	LocalDevicePtr	local,
	int		flags)
{
	TsDevicePtr	tsdev = (TsDevicePtr) local->private;
	int i;

	TsUSB_DeviceControl(local->dev, DEVICE_OFF);

	for ( i=0; i<TSUSB_MINORS; i++ ) {
		if (tsdev->priv[i]) {
			if (tsdev->priv[i]->input_dev) {
				xfree(tsdev->priv[i]->input_dev);
			 	tsdev->priv[i]->input_dev = NULL;
			}
			xfree(tsdev->priv[i]);
			tsdev->priv[i] = NULL;
		}
	}

	xfree(tsdev->devfs_name);
	xfree(tsdev);
	xfree(local->name);
	xfree(local);

	xf86DeleteInput(local, 0);

}

static
InputDriverRec TSUSB_DRIVER_RECORD = { 1, "tsusb", NULL, TsUSB_Init, TsUSB_Uninit, NULL, 0 };

static pointer
TsUSB_Plug(
	pointer	module,
	pointer	options,
	int	*errmaj,
	int	*errmin)
{
	xf86AddInputDriver(&TSUSB_DRIVER_RECORD, module, 0);
	return module;
}

static void
TsUSB_Unplug(pointer p)
{
}

static XF86ModuleVersionInfo version_rec = {
  "tsusb",
  "CometrueWorks",
  MODINFOSTRING1,
  MODINFOSTRING2,
  XORG_VERSION_CURRENT,
  TSINI_VERSION_D1,TSINI_VERSION_D2,TSINI_VERSION_D3,
  ABI_CLASS_XINPUT,
  ABI_XINPUT_VERSION,
  MOD_CLASS_XINPUT,
  { 0, 0, 0, 0 }
};

XF86ModuleData tsusbModuleData = { &version_rec, TsUSB_Plug, TsUSB_Unplug };
