/*
 * calibrate.c
 *
 */

#include <gtk/gtk.h>

#include <stdio.h>
#include <string.h>

#include "interface.h"
#include "support.h"
#include "tsequ.h"
#include "tsusbhid.h"
#include "tsevdev.h"
#include "tshiddev.h"
#include "calibrate.h"
#include "pcsound.h"

#undef	TS_CALIB_TEST_AUTO_CALCEL

#define	TS_CALIB_PROC_CLOCK	50

#define	TS_CALIB_IDEL_CLOCK	(1000/TS_CALIB_PROC_CLOCK)
#define	TS_CALIB_IDEL_TIMEOUT	15
#define	TS_CALIB_IDEL_TIMEBAR	20

#define	TS_CALIB_HOLD_CLOCK	(500/TS_CALIB_PROC_CLOCK)
#define	TS_CALIB_HOLD_SAMPLES	4
#define	TS_CALIB_HOLD_RANGE	4

#define	TS_CALIB_MARK_WW	16
#define	TS_CALIB_MARK_HH	16

#define	TS_CALIB_POINT_MIN	1
#define	TS_CALIB_POINT_MAX	8

enum {
	CALIB_STATE_START,
	CALIB_STATE_WAIT_START,
	CALIB_STATE_WAIT_TOUCH,
	CLAIB_STATE_CHECK_HOLD,
	CLAIB_STATE_WAIT_LIFT_OFF,
	CLAIB_STATE_TEST
	};

static gboolean calibrate_complete;

static GtkWidget *window_main;
static GtkWidget *window_calibrate;

static gint gnScreenWW;
static gint gnScreenHH;
static gint gnScreenCW;
static gint gnScreenCH;

static struct tsusb_feature_report ini_feature;
static struct tsusb_feature_report now_feature;

static gint	raw_x;
static gint	raw_y;
static gint	raw_count;
static gint	sample_count;
static gint	sample_xx[TS_CALIB_HOLD_SAMPLES];
static gint	sample_yy[TS_CALIB_HOLD_SAMPLES];

static gint	gnRead_XX[TS_CALIB_POINT_MAX+1];
static gint	gnRead_YY[TS_CALIB_POINT_MAX+1];

static gint	gnNewCalibrateX1;
static gint	gnNewCalibrateX2;
static gint	gnNewCalibrateX3;
static gint	gnNewCalibrateY1;
static gint	gnNewCalibrateY2;
static gint	gnNewCalibrateY3;
static gint	gnNewPanelDirection;

static gint calibrate_devnum = -1;

static gint calibrate_mode = TSEQU_CALIBRATION_4POINT_MODE1;

static gint calibrate_proc_timer = 0;
static gint calibrate_proc_state;
static gint calibrate_proc_count;
static gint calibrate_proc_point;

static gint calibrate_hold_clock;

static gint calibrate_idel_state;
static gint calibrate_idel_clock;
static gint calibrate_idel_remain;
static gint calibrate_idel_count;

static gint mark_init_count;
static gboolean mark_initialized[TS_CALIB_POINT_MAX];
static gboolean mark_state[TS_CALIB_POINT_MAX];

static gchar mbs1[255];
static gchar mbs2[255];

#define SET_MBS1(args...) snprintf (mbs1,sizeof(mbs1),args);
#define SET_MBS2(args...) snprintf (mbs2,sizeof(mbs2),args);

static void
beep_touch(void)
{
	pcsound_output(100,2468,20);
}

static void
beep_error(void)
{
	pcsound_output(100,3000,50);
}

static void
beep_done(void)
{
	static pcsound_table beep_done_data [] = {
	{100,2000,50},
	{100,2300,50},
	{100,2600,50},
	{0}
	};
	pcsound_proc_start(beep_done_data);
}

static void
beep_note(void)
{
	static pcsound_table beep_note_data [] = {
	{100,2000,50},
	{100,3000,50},
	{100,2000,50},
	{0}
	};
	pcsound_proc_start(beep_note_data);
}

static void
calibrate_msgbox_error (GtkWidget *window, const gchar *msg,  const gchar *subtext)
{
	GtkWidget *dialog;

	if ( window == window_calibrate ) {
		gtk_window_unfullscreen (GTK_WINDOW (window_calibrate));
		gtk_widget_hide (window_calibrate);
	}
	dialog = gtk_message_dialog_new (
			GTK_WINDOW (window_calibrate),
			GTK_DIALOG_MODAL | GTK_DIALOG_DESTROY_WITH_PARENT,
			GTK_MESSAGE_ERROR,
			GTK_BUTTONS_OK,
			"%s\n%s",msg,subtext);
	gtk_window_set_position (GTK_WINDOW (dialog), GTK_WIN_POS_CENTER);

	beep_error ();

	gtk_dialog_run (GTK_DIALOG (dialog));
	gtk_widget_destroy (dialog);

}

static GtkWidget *
lookup_drawingarea (gint point)
{
	GtkWidget *drawingarea = NULL;

	switch (point)
	{
	case 1:
		drawingarea = lookup_widget (window_calibrate,  "drawingarea1");
		break;
	case 2:
		drawingarea = lookup_widget (window_calibrate,  "drawingarea2");
		break;
	case 3:
		drawingarea = lookup_widget (window_calibrate,  "drawingarea3");
		break;
	case 4:
		drawingarea = lookup_widget (window_calibrate,  "drawingarea4");
		break;
	case 5:
		drawingarea = lookup_widget (window_calibrate,  "drawingarea5");
		break;
	case 6:
		drawingarea = lookup_widget (window_calibrate,  "drawingarea6");
		break;
	case 7:
		drawingarea = lookup_widget (window_calibrate,  "drawingarea7");
		break;
	case 8:
		drawingarea = lookup_widget (window_calibrate,  "drawingarea8");
		break;
	default:
		printf ("lookup_drawingarea: invalid point number (%d)\n",point);
		perror("");
		return;
	}

	return drawingarea;
}

static void
show_cursor (GdkWindow *window)
{
	gdk_window_set_cursor(window, NULL);

}

static void
hide_cursor (GdkWindow *window)
{
	guchar bits[] = { 0 };
	GdkColor color = {0,0,0,0};
	GdkPixmap *pixmap;
	GdkCursor *cursor;
	pixmap = gdk_bitmap_create_from_data(NULL, bits, 1, 1);
	cursor = gdk_cursor_new_from_pixmap (pixmap, pixmap, &color,  &color, 0, 0);
	gdk_window_set_cursor(window, cursor);

}

static void
hide_drawingareas_cursor (void)
{
	gint	point;
	GtkWidget *drawingarea;
	for ( point=TS_CALIB_POINT_MIN; point<=TS_CALIB_POINT_MAX; point++ ) {
		drawingarea = lookup_drawingarea(point);
		hide_cursor(drawingarea->window);
	}
}

static void
show_drawingareas_cursor (void)
{
	gint	point;
	GtkWidget *drawingarea;
	for ( point=TS_CALIB_POINT_MIN; point<=TS_CALIB_POINT_MAX; point++ ) {
		drawingarea = lookup_drawingarea(point);
		show_cursor(drawingarea->window);
	}
}

static void
calibrate_idel_update (void)
{
	GtkWidget *drawingarea = lookup_widget (window_calibrate,  "drawingarea_idel");
	GdkRectangle rect = {0,0,drawingarea->allocation.width,drawingarea->allocation.height};
	gdk_window_invalidate_rect(drawingarea->window,&rect,0);
}

static void
calibrate_idel_string (void)
{
	GtkWidget *label = lookup_widget (window_calibrate,  "label_idel");;
	char text[256];
	switch (calibrate_idel_state)
	{
	case 0:
		snprintf (text,sizeof(text),"----------------------------------");
		break;
	case 1:
		if (calibrate_idel_remain)	snprintf (text,sizeof(text),"Auto Cancel after %d second%s",calibrate_idel_remain,calibrate_idel_remain>1?"s":"");
		else				snprintf (text,sizeof(text),"- - -  A U T O  C A N C E L  - - -");
		break;
	default:
		snprintf (text,sizeof(text),"*** invalid state %d ***", calibrate_idel_state);
		break;
	}
	gtk_label_set ( GTK_LABEL(label), text );
}

static void
calibrate_init_marks (void)
{
	gint	point;

	mark_init_count = 0;

	for ( point=TS_CALIB_POINT_MIN; point<=TS_CALIB_POINT_MAX; point++ ) {
		mark_state[point] = FALSE;
		mark_initialized[point] = FALSE;
	}
}

static void
calibrate_mark_show (gint point)
{
	GtkWidget *drawingarea = lookup_drawingarea (point);
	GdkRectangle rect = {0,0,drawingarea->allocation.width,drawingarea->allocation.height};
	gdk_window_invalidate_rect (drawingarea->window,&rect,0);
}

static void
calibrate_mark_on (gint point)
{
	mark_state[point] = TRUE;
	calibrate_mark_show (point);
}

static void
calibrate_mark_off (gint point)
{
	mark_state[point] = FALSE;
	calibrate_mark_show (point);
}

static void
calibrate_help_string (gint point)
{

	#define TS_CALIB_HELP_SPAN_LEAD	"<span foreground=\"#008800\">"
	#define TS_CALIB_HELP_SPAN_END	"</span>"
	#define TS_CALIB_HELP_BASE	TS_CALIB_HELP_SPAN_LEAD "Touch on the center of cross that\nshow on the %s of screen.\n" TS_CALIB_HELP_SPAN_END

	GtkWidget *label = lookup_widget (window_calibrate,  "label_help");
	char text[512];

	switch (point)
	{
	case -1:
		snprintf (text,sizeof(text),
			"<span foreground=\"#FF0000\" background=\"#FFFF00\" size=\"x-large\">"
			"<b>  Lift off to start calibration  </b>"
			"</span>"
			);
		break;
	case 0:
		snprintf (text,sizeof(text), TS_CALIB_HELP_SPAN_LEAD "OK! Lift off to next step.\n" TS_CALIB_HELP_SPAN_END);
		break;
	case 1:
		snprintf (text,sizeof(text),TS_CALIB_HELP_BASE,"LEFT-TOP");
		break;
	case 2:
		snprintf (text,sizeof(text),TS_CALIB_HELP_BASE,"RIGHT-TOP");
		break;
	case 3:
		snprintf (text,sizeof(text),TS_CALIB_HELP_BASE,"RIGHT-BOTTOM");
		break;
	case 4:
		snprintf (text,sizeof(text),TS_CALIB_HELP_BASE,"LEFT-BOTTOM");
		break;
	case 5:
		snprintf (text,sizeof(text),TS_CALIB_HELP_BASE,"TOP");
		break;
	case 6:
		snprintf (text,sizeof(text),TS_CALIB_HELP_BASE,"RIGHT");
		break;
	case 7:
		snprintf (text,sizeof(text),TS_CALIB_HELP_BASE,"BOTTOM");
		break;
	case 8:
		snprintf (text,sizeof(text),TS_CALIB_HELP_BASE,"LEFT");
		break;
	case 255:
		snprintf (text,sizeof(text),
				"<span foreground=\"#008800\">"
				"<b>Calibration is complete!\n</b>"
				"To test,\ntouch on and check cursor position.\n"
				"[Save] to save new data and exit.\n"
				"[Redo] to run calibration again.\n"
				"[Cancel] to restore old data and exit."
				"</span>"
				);
		break;
	default:
		printf ("calibrate_help_string: invalid point number (%d)\n",point);
		perror ("");
		return;
	}
 	gtk_label_set ( GTK_LABEL(label), text );
	gtk_label_set_use_markup (GTK_LABEL (label), TRUE);
}

static void
PanelDirectionString (char *S)
{

	gint wPX1, wPX2, wPX3, wPX4;
	gint wPY1, wPY2, wPY3, wPY4;

	switch (calibrate_mode)
	{
	case TSEQU_CALIBRATION_4POINT_MODE1:
		wPX1 = gnRead_XX[1];
		wPY1 = gnRead_YY[1];
		wPX2 = gnRead_XX[2];
		wPY2 = gnRead_YY[2];
		wPX3 = gnRead_XX[3];
		wPY3 = gnRead_YY[3];
		wPX4 = gnRead_XX[4];
		wPY4 = gnRead_YY[4];
		break;
	case TSEQU_CALIBRATION_4POINT_MODE2:
		wPX1 = gnRead_XX[8];
		wPY1 = gnRead_YY[5];
		wPX2 = gnRead_XX[6];
		wPY2 = gnRead_YY[5];
		wPX3 = gnRead_XX[6];
		wPY3 = gnRead_YY[7];
		wPX4 = gnRead_XX[8];
		wPY4 = gnRead_YY[7];
		break;
	}
	S[0] = (wPX1 < TSEQU_ADCIN_PDC)? '0': '1';
	S[1] = (wPY1 < TSEQU_ADCIN_PDC)? '0': '1';
	S[2] = (wPX2 < TSEQU_ADCIN_PDC)? '0': '1';
	S[3] = (wPY2 < TSEQU_ADCIN_PDC)? '0': '1';
	S[4] = (wPX3 < TSEQU_ADCIN_PDC)? '0': '1';
	S[5] = (wPY3 < TSEQU_ADCIN_PDC)? '0': '1';
	S[6] = (wPX4 < TSEQU_ADCIN_PDC)? '0': '1';
	S[7] = (wPY4 < TSEQU_ADCIN_PDC)? '0': '1';
	S[8] = '\0';

}

static gint
PanelDirectionCode(char *strPD)
{
		const char* szPD0 = "00101101";
		const char* szPD1 = "10000111";
		const char* szPD2 = "01111000";
		const char* szPD3 = "11010010";

		int nPD  = 255;
		int slen = strlen(szPD0);

		if	( strncmp (strPD,szPD0,slen) == 0 ) nPD = 0;
		else if	( strncmp (strPD,szPD1,slen) == 0 ) nPD = 1;
		else if	( strncmp (strPD,szPD2,slen) == 0 ) nPD = 2;
		else if	( strncmp (strPD,szPD3,slen) == 0 ) nPD = 3;

		return (nPD);
}

static gboolean
PanelDirectionCheck(void)
{

	gchar S1[] = {0,0,0,0,0,0,0,0,0};
	gchar S2[] = {0,0,0,0,0,0,0,0,0};

	gint  PD;
	gint  temp;

	PanelDirectionString(S1);
	PD = PanelDirectionCode(S1);

	if ( PD == 255 ) {
		switch (calibrate_mode)
		{
			#define SWAP(t,a,b) t=a; a=b; b=t;
			case TSEQU_CALIBRATION_4POINT_MODE1:
				SWAP( temp, gnRead_XX[1], gnRead_YY[1] );
				SWAP( temp, gnRead_XX[2], gnRead_YY[2] );
				SWAP( temp, gnRead_XX[3], gnRead_YY[3] );
				SWAP( temp, gnRead_XX[4], gnRead_YY[4] );
				break;
			case TSEQU_CALIBRATION_4POINT_MODE2:
				SWAP( temp, gnRead_XX[5], gnRead_YY[5] );
				SWAP( temp, gnRead_XX[6], gnRead_YY[6] );
				SWAP( temp, gnRead_XX[7], gnRead_YY[7] );
				SWAP( temp, gnRead_XX[8], gnRead_YY[8] );
				break;
			default:
				return FALSE;
		}
		PanelDirectionString(S2);
		PD = PanelDirectionCode(S2);
		if ( PD == 255 ) {
			SET_MBS1 ("Incorrect Panel Direction! ( %s - %s )", S1, S2 );
			SET_MBS2 ("It may be caused by the Pinout of Touch Panel is not supported or touched on incorrect position during the calibration." );
			calibrate_msgbox_error ( window_calibrate, mbs1, mbs2 );
			return FALSE;
		}
		PD = PD + 4;
	}

	gnNewPanelDirection = PD;

	return TRUE;

}

static gboolean
CalibrationResultCheck(void)
{
	gchar s[5] = {'\0',/*...*/};

	if ( (gnNewCalibrateX2-gnNewCalibrateX1) < 256 ) strncat(s, "1", sizeof(s));
	if ( (gnNewCalibrateX3-gnNewCalibrateX2) < 256 ) strncat(s, "2", sizeof(s));
	if ( (gnNewCalibrateY2-gnNewCalibrateY1) < 256 ) strncat(s, "3", sizeof(s));
	if ( (gnNewCalibrateY3-gnNewCalibrateY2) < 256 ) strncat(s, "4", sizeof(s));
	if (strlen(s)) {
		SET_MBS1 ( "Invalid result of calibration ( %s )", s );
		SET_MBS2 ( "It may be caused by touched on incorrect position during the calibration." );
		calibrate_msgbox_error ( window_calibrate, mbs1, mbs2 );
		return FALSE;
	}
	return TRUE;
}

static void
Calibrate4PointMode1(void)
{

	int wPX1, wPX2, wPX3, wPX4;
	int wPY1, wPY2, wPY3, wPY4;

	int X, Y;

	if ( gnNewPanelDirection & 2) {
		gnRead_YY[1] = TSEQU_ADCIN_MAX - gnRead_YY[1];
		gnRead_YY[2] = TSEQU_ADCIN_MAX - gnRead_YY[2];
		gnRead_YY[3] = TSEQU_ADCIN_MAX - gnRead_YY[3];
		gnRead_YY[4] = TSEQU_ADCIN_MAX - gnRead_YY[4];
	}

	if ( gnNewPanelDirection & 1) {
		gnRead_XX[1] = TSEQU_ADCIN_MAX - gnRead_XX[1];
		gnRead_XX[2] = TSEQU_ADCIN_MAX - gnRead_XX[2];
		gnRead_XX[3] = TSEQU_ADCIN_MAX - gnRead_XX[3];
		gnRead_XX[4] = TSEQU_ADCIN_MAX - gnRead_XX[4];
	}

	wPX1 = gnRead_XX[1];
	wPY1 = gnRead_YY[1];
	wPX2 = gnRead_XX[2];
	wPY2 = gnRead_YY[2];
	wPX3 = gnRead_XX[3];
	wPY3 = gnRead_YY[3];
	wPX4 = gnRead_XX[4];
	wPY4 = gnRead_YY[4];

	gnNewCalibrateX1 = (wPX1 + wPX4) / 2;
	gnNewCalibrateX3 = (wPX2 + wPX3) / 2;
	gnNewCalibrateY1 = (wPY1 + wPY2) / 2;
	gnNewCalibrateY3 = (wPY3 + wPY4) / 2;

	gnNewCalibrateX2 = gnNewCalibrateX1 + (gnNewCalibrateX3 - gnNewCalibrateX1) / 2;
	gnNewCalibrateY2 = gnNewCalibrateY1 + (gnNewCalibrateY3 - gnNewCalibrateY1) / 2;

	X = (gnNewCalibrateX3 - gnNewCalibrateX1 + 1) * (TS_CALIB_MARK_WW+TS_CALIB_BOX_BORDER) / (gnScreenWW - ((TS_CALIB_MARK_WW+TS_CALIB_BOX_BORDER) * 2));
	Y = (gnNewCalibrateY3 - gnNewCalibrateY1 + 1) * (TS_CALIB_MARK_HH+TS_CALIB_BOX_BORDER) / (gnScreenHH - ((TS_CALIB_MARK_HH+TS_CALIB_BOX_BORDER) * 2));

	gnNewCalibrateX1 = gnNewCalibrateX1 - X;
	gnNewCalibrateX3 = gnNewCalibrateX3 + X;

	gnNewCalibrateY1 = gnNewCalibrateY1 - Y;
	gnNewCalibrateY3 = gnNewCalibrateY3 + Y;

	if ( gnNewCalibrateX1 < TSEQU_ADCIN_MIN ) gnNewCalibrateX1 = TSEQU_ADCIN_MIN;
	if ( gnNewCalibrateX3 > TSEQU_ADCIN_MAX ) gnNewCalibrateX3 = TSEQU_ADCIN_MAX;
	if ( gnNewCalibrateY1 < TSEQU_ADCIN_MIN ) gnNewCalibrateY1 = TSEQU_ADCIN_MIN;
	if ( gnNewCalibrateY3 > TSEQU_ADCIN_MAX ) gnNewCalibrateY3 = TSEQU_ADCIN_MAX;

}

static void
Calibrate4PointMode2(void)
{

	int wPX1, wPX2, wPX3, wPX4, wPX5, wPX6, wPX7, wPX8;
	int wPY1, wPY2, wPY3, wPY4, wPY5, wPY6, wPY7, wPY8;

	int YU, YD, XL, XR;

	if (gnNewPanelDirection & 2) {
		gnRead_YY[5] = TSEQU_ADCIN_MAX - gnRead_YY[5];
		gnRead_YY[6] = TSEQU_ADCIN_MAX - gnRead_YY[6];
		gnRead_YY[7] = TSEQU_ADCIN_MAX - gnRead_YY[7];
		gnRead_YY[8] = TSEQU_ADCIN_MAX - gnRead_YY[8];
	}

	if (gnNewPanelDirection & 1) {
		gnRead_XX[5] = TSEQU_ADCIN_MAX - gnRead_XX[5];
		gnRead_XX[6] = TSEQU_ADCIN_MAX - gnRead_XX[6];
		gnRead_XX[7] = TSEQU_ADCIN_MAX - gnRead_XX[7];
		gnRead_XX[8] = TSEQU_ADCIN_MAX - gnRead_XX[8];
	}

	wPX1 = gnRead_XX[8];
	wPY1 = gnRead_YY[5];
	wPX2 = gnRead_XX[5];
	wPY2 = gnRead_YY[5];
	wPX3 = gnRead_XX[6];
	wPY3 = gnRead_YY[5];
	wPX4 = gnRead_XX[6];
	wPY4 = gnRead_YY[6];
	wPX5 = gnRead_XX[6];
	wPY5 = gnRead_YY[7];
	wPX6 = gnRead_XX[7];
	wPY6 = gnRead_YY[7];
	wPX7 = gnRead_XX[8];
	wPY7 = gnRead_YY[7];
	wPX8 = gnRead_XX[8];
	wPY8 = gnRead_YY[8];

	gnNewCalibrateX1 = (wPX1 + wPX7 + wPX8) / 3;
	gnNewCalibrateX3 = (wPX3 + wPX4 + wPX5) / 3;
	gnNewCalibrateY1 = (wPY1 + wPY2 + wPY3) / 3;
	gnNewCalibrateY3 = (wPY5 + wPY6 + wPY7) / 3;

	gnNewCalibrateX2 = (wPX2 + wPX6) / 2;
	gnNewCalibrateY2 = (wPY4 + wPY8) / 2;

	XL = ((gnNewCalibrateX2 - gnNewCalibrateX1 + 1) * (TS_CALIB_MARK_WW+TS_CALIB_BOX_BORDER) / (gnScreenCW-(TS_CALIB_MARK_WW+TS_CALIB_BOX_BORDER)));
	XR = ((gnNewCalibrateX3 - gnNewCalibrateX2 + 1) * (TS_CALIB_MARK_WW+TS_CALIB_BOX_BORDER) / (gnScreenCW-(TS_CALIB_MARK_WW+TS_CALIB_BOX_BORDER)));
	YU = ((gnNewCalibrateY2 - gnNewCalibrateY1 + 1) * (TS_CALIB_MARK_HH+TS_CALIB_BOX_BORDER) / (gnScreenCH-(TS_CALIB_MARK_HH+TS_CALIB_BOX_BORDER)));
	YD = ((gnNewCalibrateY3 - gnNewCalibrateY2 + 1) * (TS_CALIB_MARK_HH+TS_CALIB_BOX_BORDER) / (gnScreenCH-(TS_CALIB_MARK_HH+TS_CALIB_BOX_BORDER)));

	gnNewCalibrateX1 = gnNewCalibrateX1 - XL;
	gnNewCalibrateX3 = gnNewCalibrateX3 + XR;

	gnNewCalibrateY1 = gnNewCalibrateY1 - YU;
	gnNewCalibrateY3 = gnNewCalibrateY3 + YD;

	if ( gnNewCalibrateX1 < TSEQU_ADCIN_MIN ) gnNewCalibrateX1 = TSEQU_ADCIN_MIN;
	if ( gnNewCalibrateX3 > TSEQU_ADCIN_MAX ) gnNewCalibrateX3 = TSEQU_ADCIN_MAX;
	if ( gnNewCalibrateY1 < TSEQU_ADCIN_MIN ) gnNewCalibrateY1 = TSEQU_ADCIN_MIN;
	if ( gnNewCalibrateY3 > TSEQU_ADCIN_MAX ) gnNewCalibrateY3 = TSEQU_ADCIN_MAX;

}

static gboolean
calibrate_set_new_data(void)
{
	gint ret;

	now_feature.feature_id		= TSUSB_FEATURE_ID_SAVE;
	now_feature.calibration_x1	= gnNewCalibrateX1;
	now_feature.calibration_x2	= gnNewCalibrateX2;
	now_feature.calibration_x3	= gnNewCalibrateX3;
	now_feature.calibration_y1	= gnNewCalibrateY1;
	now_feature.calibration_y2	= gnNewCalibrateY2;
	now_feature.calibration_y3	= gnNewCalibrateY3;
	now_feature.panel_direction	= gnNewPanelDirection;

	ret = tshiddev_set_feature(&now_feature);

	if ( ret < 0 ) {
		SET_MBS1 ( "Failed to set new calibration data to device!" );
		SET_MBS2 ( "device '" TSEVDEV_DEVICE_NAME "%d', error %d", calibrate_devnum, ret );
		calibrate_msgbox_error ( window_calibrate, mbs1, mbs2 );
		return FALSE;
	}
	return TRUE;
}

static gboolean
calibrate_restore_old_data(void)
{
	gint ret;

	now_feature.feature_id		= TSUSB_FEATURE_ID_SAVE;
	now_feature.calibration_x1	= ini_feature.calibration_x1;
	now_feature.calibration_x2	= ini_feature.calibration_x2;
	now_feature.calibration_x3	= ini_feature.calibration_x3;
	now_feature.calibration_y1	= ini_feature.calibration_y1;
	now_feature.calibration_y2	= ini_feature.calibration_y2;
	now_feature.calibration_y3	= ini_feature.calibration_y3;
	now_feature.panel_direction	= ini_feature.panel_direction;

	ret = tshiddev_set_feature(&now_feature);

	if ( ret < 0 ) {
		SET_MBS1 ( "Failed to restore old calibration data to device!" );
		SET_MBS2 ( "device '" TSEVDEV_DEVICE_NAME "%d', error %d", calibrate_devnum, ret );
		calibrate_msgbox_error ( window_calibrate, mbs1, mbs2 );
		return FALSE;
	}
	
	return TRUE;
	
}

static void
calibrate_proc_initialize(void)
{
	gint ret;

	ret = tshiddev_disable();

	if ( ret < 0 ) {
		printf ("calibrate_proc_complete: Failed to auto disable the device! ( devnum=%d, error=%d )\n",calibrate_devnum,ret);
		perror("");
	}

	calibrate_complete = FALSE;

	gtk_widget_show (lookup_widget (window_calibrate, "label_idel"));
	gtk_widget_show (lookup_widget (window_calibrate, "drawingarea_idel"));
	gtk_widget_show (lookup_widget (window_calibrate, "button_break"));

	gtk_widget_hide (lookup_widget (window_calibrate, "button_save"));
	gtk_widget_hide (lookup_widget (window_calibrate, "button_redo"));
	gtk_widget_hide (lookup_widget (window_calibrate, "button_cancel"));

	hide_drawingareas_cursor();

}

static void
calibrate_proc_complete(void)
{

	gint ret;

	ret = tshiddev_enable();

	if ( ret < 0 ) {
		printf ("calibrate_proc_complete: Failed to auto enable the device! ( devnum=%d, error=%d )\n",calibrate_devnum,ret);
		perror("");
	}

	calibrate_complete = TRUE;

	gtk_widget_hide (lookup_widget (window_calibrate, "button_break"));

	gtk_widget_show (lookup_widget (window_calibrate, "button_save"));
	gtk_widget_show (lookup_widget (window_calibrate, "button_redo"));
	gtk_widget_show (lookup_widget (window_calibrate, "button_cancel"));

}

static gint
calibrate_proc(gpointer data)
{
	gint i;
	gint ret;
	gint idel_state = calibrate_idel_state;
	gint idel_count;

	ret = tshiddev_get_feature(&now_feature);

	switch (calibrate_proc_state)
	{
	case CALIB_STATE_START:
		if ( now_feature.raw_touch ) {
			idel_state = 1;
			calibrate_help_string(-1);
			calibrate_proc_state = CALIB_STATE_WAIT_START;
			beep_note ();
		}
		else {
			calibrate_proc_point = ((calibrate_mode==TSEQU_CALIBRATION_4POINT_MODE1)? 1: 5);
			calibrate_proc_count = 4;
			calibrate_mark_on(calibrate_proc_point);
			calibrate_help_string(calibrate_proc_point);
			idel_state = 0;
			calibrate_proc_state = CALIB_STATE_WAIT_TOUCH;
		}
		break;
	case CALIB_STATE_WAIT_START:
		if ( !now_feature.raw_touch ) {
			calibrate_proc_state = CALIB_STATE_START;
		}
		break;
	case CALIB_STATE_WAIT_TOUCH:
		if ( now_feature.raw_touch ) {
			idel_state = 0;
			raw_x = now_feature.raw_x;
			raw_y = now_feature.raw_y;
			raw_count = now_feature.raw_count;
			sample_count = 0;
			calibrate_hold_clock = TS_CALIB_HOLD_CLOCK;
			calibrate_proc_state = CLAIB_STATE_CHECK_HOLD;
			beep_touch();
		}
		else {
			idel_state = 1;
		}
		break;
	case CLAIB_STATE_CHECK_HOLD:
		if ( !now_feature.raw_touch ) {
			calibrate_proc_state = CALIB_STATE_WAIT_TOUCH;
		}
		else {
			if ( calibrate_hold_clock-- ) {
				if ( sample_count < TS_CALIB_HOLD_SAMPLES ) {
					if ( raw_count != now_feature.raw_count ) {
						raw_count = now_feature.raw_count;
						if (	abs(raw_x-now_feature.raw_x) > TS_CALIB_HOLD_RANGE ||
							abs(raw_y-now_feature.raw_y) > TS_CALIB_HOLD_RANGE ) {
							raw_x = now_feature.raw_x;
							raw_y = now_feature.raw_y;
							sample_count = 0;
						}
						else {
							sample_xx[sample_count] = now_feature.raw_x;
							sample_yy[sample_count] = now_feature.raw_y;
							sample_count++;
						}
					}
				}
			}
			else {
				if ( sample_count != TS_CALIB_HOLD_SAMPLES ) {
					calibrate_hold_clock = TS_CALIB_HOLD_CLOCK;
				}
				else {
					raw_x = 0;
					raw_y =  0;
					for ( i=0; i<sample_count; i++ ) {
						raw_x += sample_xx[i];
						raw_y += sample_yy[i];
					}
					gnRead_XX[calibrate_proc_point] = raw_x / sample_count;
					gnRead_YY[calibrate_proc_point] = raw_y / sample_count;
					calibrate_mark_off(calibrate_proc_point);
					calibrate_proc_state = CLAIB_STATE_WAIT_LIFT_OFF;
					beep_done();
				}
			}
		}
		break;
	case CLAIB_STATE_WAIT_LIFT_OFF:
		if ( now_feature.raw_touch ) {
			calibrate_help_string(0);
		}
		else {
			if (--calibrate_proc_count) {
				calibrate_proc_point++;
				calibrate_mark_on(calibrate_proc_point);
				calibrate_help_string(calibrate_proc_point);
				calibrate_proc_state = CALIB_STATE_WAIT_TOUCH;
			}
			else {

				if (!PanelDirectionCheck()) {
					calibrate_end(CALIB_RESULT_CANCEL);
					return 0;
				}

				if 	( calibrate_mode == TSEQU_CALIBRATION_4POINT_MODE1 ) Calibrate4PointMode1();
				else if ( calibrate_mode == TSEQU_CALIBRATION_4POINT_MODE2 ) Calibrate4PointMode2();

				if (!CalibrationResultCheck()) {
					calibrate_end(CALIB_RESULT_CANCEL);
					return 0;
				}

				if (!calibrate_set_new_data()) {
					calibrate_end(CALIB_RESULT_CANCEL);
					return 0;
				}

				calibrate_proc_complete();

				calibrate_help_string(255);
				show_drawingareas_cursor();
				
				#ifndef	TS_CALIB_TEST_AUTO_CALCEL
					gtk_widget_hide (lookup_widget (window_calibrate,  "label_idel"));
					gtk_widget_hide (lookup_widget (window_calibrate,  "drawingarea_idel"));
				#endif
				
				calibrate_proc_state = CLAIB_STATE_TEST;
			}
		}
		break;
	case CLAIB_STATE_TEST:
		#ifdef	TS_CALIB_TEST_AUTO_CALCEL
			idel_state = now_feature.raw_touch? 0: 1;
		#endif
		break;
	default:
		break;
	}

	if ( calibrate_idel_state ) {
		if (!calibrate_idel_clock--) {
			calibrate_idel_clock = TS_CALIB_IDEL_CLOCK;
			if (!calibrate_idel_remain--) {
				calibrate_end(CALIB_RESULT_CANCEL);
				return 0;
			}
			calibrate_idel_string();
			idel_count = (TS_CALIB_IDEL_TIMEOUT-calibrate_idel_remain) * TS_CALIB_IDEL_TIMEBAR / TS_CALIB_IDEL_TIMEOUT;
			if ( calibrate_idel_count != idel_count ) {
				calibrate_idel_count = idel_count;
				calibrate_idel_update();
			}
		}
	}
	else {
		if ( calibrate_idel_remain < TS_CALIB_IDEL_TIMEOUT ) {
			calibrate_idel_remain = TS_CALIB_IDEL_TIMEOUT;
			calibrate_idel_clock = TS_CALIB_IDEL_CLOCK;
			calibrate_idel_count = 0;
			calibrate_idel_string();
			calibrate_idel_update();
		}
	}
	if ( calibrate_idel_state != idel_state ) {
		calibrate_idel_state = idel_state;
		calibrate_idel_string();
	}

	return 1;
}

static void
calibrate_proc_start(void)
{

	calibrate_proc_state  = CALIB_STATE_START;
	calibrate_idel_state  = 0;
	calibrate_idel_clock  = TS_CALIB_IDEL_CLOCK;
	calibrate_idel_remain = TS_CALIB_IDEL_TIMEOUT;
	calibrate_idel_count  = 0;
	calibrate_idel_update();

	calibrate_proc_timer = g_timeout_add(
				TS_CALIB_PROC_CLOCK,
				calibrate_proc,
				NULL
				);
}

static void
calibrate_proc_end(void)
{
	if (calibrate_proc_timer) {
		g_source_remove(calibrate_proc_timer);
		calibrate_proc_timer = 0;
	}
}

void
calibrate_draw_idel (GtkWidget *drawingarea)
{
	GdkGC *gc = drawingarea->style->black_gc;
	GdkColormap *colormap = gtk_widget_get_colormap(drawingarea);
	GdkColor color_fore = {0,0,0,65535};
	GdkColor color_black = {0,0,0,0};
	gint i;
	gint x, w, h, t, z, loop;

	gdk_rgb_find_color(colormap,&color_fore);
	gdk_rgb_find_color(colormap,&color_black);

	t = TS_CALIB_IDEL_TIMEBAR;
	x = 10;
	w = drawingarea->allocation.width - x * 2;
	h  = drawingarea->allocation.height;
	z = w / ( t *2 - 1);
	x = x + ( ( w - z * ( t *2 - 1) ) / 2 );

	for ( i=0; i<t; i++) {
		if ( i<calibrate_idel_count )	gdk_gc_set_foreground(gc,&color_fore);
		else				gdk_gc_set_foreground(gc,&color_black);
		gdk_draw_rectangle (drawingarea->window,
						gc,
						TRUE,
						x,0,
						z+(z/2),
						h);
		x = x + z * 2;
	}
	gdk_gc_set_foreground(gc,&color_black);

}

void
calibrate_draw_mark (gint point, GtkWidget *drawingarea)
{
	gint x, y, w, h;

	gint x1, y1, x2, y2;

	GdkGC *gc;
	GdkColormap *colormap = gtk_widget_get_colormap(drawingarea);
	GdkColor color_back = {0,0,0,0};
	GdkColor color_fore = {0,65535,65535,65535};
	GdkColor color_mark = {0,65535,65535,0};
	GdkColor color_marc = {0,65535,0,0};

	gc = drawingarea->style->black_gc;

	gdk_rgb_find_color(colormap,&color_back);
	gdk_rgb_find_color(colormap,&color_fore);
	gdk_rgb_find_color(colormap,&color_mark);
	gdk_rgb_find_color(colormap,&color_marc);

	if ( !mark_initialized[point] ) {

		gdk_gc_set_foreground(gc,&color_back);
		gdk_gc_set_background(gc,&color_back);
		gdk_draw_rectangle (drawingarea->window,
						gc,
						TRUE,
						0,0,
						drawingarea->allocation.width,
						drawingarea->allocation.height);
		mark_initialized[point] = TRUE;
		mark_init_count++;
		if ( mark_init_count == TS_CALIB_POINT_MAX ) {
			calibrate_proc_start();
		}
		return;
	}

	if ( mark_init_count != TS_CALIB_POINT_MAX ) {
		return;
	}

	w = drawingarea->allocation.width;
	h = drawingarea->allocation.height;

	switch (point)
	{
	case 1:
		x = TS_CALIB_MARK_WW-1;
		y = TS_CALIB_MARK_HH-1;
		break;
	case 2:
		x = w - TS_CALIB_MARK_WW;
		y = TS_CALIB_MARK_HH-1;
		break;
	case 3:
		x = w - TS_CALIB_MARK_WW;
		y = h - TS_CALIB_MARK_HH;
		break;
	case 4:
		x = TS_CALIB_MARK_WW-1;
		y = h - TS_CALIB_MARK_HH;
		break;
	case 5:
		x = w / 2;
		y = TS_CALIB_MARK_HH-1;
		break;
	case 6:
		x = w - TS_CALIB_MARK_WW;
		y = h / 2;
		break;
	case 7:
		x = w / 2;
		y = h - TS_CALIB_MARK_HH;
		break;
	case 8:
		x = TS_CALIB_MARK_WW-1;
		y = h / 2 ;
		break;
	default:
		printf ("calibrate_draw_mark: invalid point number (%d)\n",point);
		perror("");
		return;
	}

	gdk_gc_set_foreground(gc,&color_back);
	gdk_gc_set_background(gc,&color_back);
	gdk_draw_rectangle (drawingarea->window,
						gc,
						TRUE,
						0,0,
						drawingarea->allocation.width,
						drawingarea->allocation.height);
	if (mark_state[point]) {

		x1 = x - TS_CALIB_MARK_WW+1;
		y1 = y - TS_CALIB_MARK_HH+1;
		x2 = x + TS_CALIB_MARK_WW-1;
		y2 = y + TS_CALIB_MARK_HH-1;

		gdk_gc_set_foreground(gc,&color_marc);

		gdk_draw_arc(drawingarea->window,
					gc,
					TRUE,
					x-(TS_CALIB_MARK_WW*2/3),
					y-(TS_CALIB_MARK_HH*2/3),
					(TS_CALIB_MARK_WW*2/3)*2,
					(TS_CALIB_MARK_HH*2/3)*2,
					0,360*64);

		gdk_gc_set_foreground(gc,&color_mark);
		gdk_draw_rectangle (drawingarea->window, gc, TRUE, x1 , y-1, x2-x1+1, 3 );
		gdk_draw_rectangle (drawingarea->window, gc, TRUE, x-1, y1, 3, y2-y1+1 );

		gdk_gc_set_foreground(gc,&color_back);
		gdk_draw_line(drawingarea->window, gc,  x1+1,y,x2-1,y);
		gdk_draw_line(drawingarea->window, gc,  x,y1+1,x,y2-1);

	}

}

void
calibrate_start(void)
{
	gint ret;

	calibrate_devnum = device_treeview_selected_devnum();

	if ( calibrate_devnum < 0 ) {
		SET_MBS1 ( "calibrate_start: invalid device number %d", calibrate_devnum );
		SET_MBS2 ( "Select a connected device from the list before Start." );
		calibrate_msgbox_error ( window_main, mbs1, mbs2 );
		return;
	}

        calibrate_mode = options_now_calibration_mode();

	ret = tshiddev_open(calibrate_devnum);
	if ( ret < 0 ) {
		SET_MBS1 ( "calibrate_start: failed to open device" );
		SET_MBS2 ( "device '" TSEVDEV_DEVICE_NAME "%d', error %d", calibrate_devnum, ret );
		calibrate_msgbox_error ( window_main, mbs1, mbs2 );
		return;
	}

	ret = tshiddev_get_feature(&ini_feature);
	if ( ret < 0 ) {
		SET_MBS1 ( "calibrate_start: failed to get feature from device" );
		SET_MBS2 ( "device '" TSEVDEV_DEVICE_NAME "%d', error %d", calibrate_devnum, ret );
		calibrate_msgbox_error ( window_main, mbs1, mbs2 );
		tshiddev_close();
		return;
	}

	calibrate_init_marks();

	gtk_window_set_modal (GTK_WINDOW (window_calibrate), TRUE);
	gtk_window_fullscreen (GTK_WINDOW (window_calibrate));

	gtk_widget_hide (window_main);
	gtk_widget_show (window_calibrate);

	calibrate_proc_initialize();

}

void
calibrate_end(CalibrateResult calibrate_result)
{

	if ( calibrate_complete ) {
		if ( calibrate_result ==  CALIB_RESULT_ESCAPE  ) return;
		if ( calibrate_result !=  CALIB_RESULT_SAVE  &&
		     calibrate_result !=  CALIB_RESULT_DESTORY  ) {
			calibrate_restore_old_data();
		}
	}
	else {
		gint ret;
		ret = tshiddev_enable();
		if ( ret < 0 ) {
			printf ("calibrate_end: Failed to auto enable the device! ( devnum=%d, error=%d )\n",calibrate_devnum,ret);
			perror("");
		}
	}

	calibrate_proc_end();

	if ( calibrate_result ==  CALIB_RESULT_REDO  ) {
		calibrate_proc_initialize();
		calibrate_proc_start();
		return;
	}

	tshiddev_close();

	gtk_window_unfullscreen (GTK_WINDOW (window_calibrate));
	gtk_widget_hide (window_calibrate);
	gtk_widget_show (window_main);
}

void
calibrate_initialize ( GtkWidget *widget_main, GtkWidget *widget_calibrate )
{
	window_main = widget_main;
	window_calibrate = widget_calibrate;

	gnScreenWW = gdk_screen_width();
	gnScreenHH = gdk_screen_height();
	gnScreenCW = gnScreenWW / 2;
	gnScreenCH = gnScreenHH / 2;

}
