/*
 * main.c
 *
 */

#include <gtk/gtk.h>

#include <stdio.h>
#include <fcntl.h>

#include "interface.h"
#include "support.h"
#include "msgbox.h"
#include "tsequ.h"
#include "tsevdev.h"
#include "options.h"
#include "calibrate.h"

#define	MIN_GTK_MAJOR_VERSION 2
#define	MIN_GTK_MINOR_VERSION 4
#define	MIN_GTK_MICRO_VERSION 0

void
set_label_string (GtkWidget *target, gboolean markup, const gchar *message_format, ...)
{
	gchar* text = NULL;
	va_list args;

	if (message_format) {
		va_start (args, message_format);
		text = g_strdup_vprintf (message_format, args);
		va_end (args);
	}

	gtk_label_set ( GTK_LABEL(target), text );
	gtk_label_set_use_markup (GTK_LABEL (target), markup);

	if (text)  g_free (text);

}

int
main (int argc, char *argv[])
{

	GtkWidget *window_main;
	GtkWidget *window_options;
	GtkWidget *window_calibrate;

	gint screen_width;
	gint screen_height;

	printf("\n" TSEQU_PROGRAM_TITLE " V" TSEQU_PROGRAM_VERSION "\n\n");

	printf("Builded with GTK+ Library Version %d.%d.%d\n",
		GTK_MAJOR_VERSION,
		GTK_MINOR_VERSION,
		GTK_MICRO_VERSION
	);

	printf("Running with GTK+ Library Version %d.%d.%d\n\n",
		gtk_major_version,
		gtk_minor_version,
		gtk_micro_version
	);

	gtk_set_locale ();
	gtk_init (&argc, &argv);

	if ( gtk_major_version < MIN_GTK_MAJOR_VERSION ||
	     gtk_minor_version < MIN_GTK_MINOR_VERSION ||
	     gtk_micro_version < MIN_GTK_MICRO_VERSION ){
		msgbox_error(	NULL,
				msgbox_string1 ("GTK+ Runtime Library Version %d.%d.%d required!",
						MIN_GTK_MAJOR_VERSION,
						MIN_GTK_MINOR_VERSION,
						MIN_GTK_MICRO_VERSION
						),
				msgbox_string2 ("You are using GTK+ Version %d.%d.%d Runtime Library.\n",
						gtk_major_version,
						gtk_minor_version,
						gtk_micro_version
						),
				GTK_BUTTONS_OK
				);
		return 1;
	}

	if (!tsevdev_open()) {
		msgbox_error(	NULL,
				msgbox_string1 ("Failed to open '%s' device.",TSEVDEV_DEVICE_NAME),
				msgbox_string2 ("Checkout TSUSB Driver Installation."),
				GTK_BUTTONS_OK
				);
		return 2;
	}

	if ( access(TSEQU_CONFIG_FILENAME,F_OK) != 0 ) {
		msgbox_error(	NULL,
				msgbox_string1 ("Failed to find '%s' file.",TSEQU_CONFIG_FILENAME),
				msgbox_string2 ("Checkout TSUSB Driver Installation."),
				GTK_BUTTONS_OK
				);
		return 3;
	}
	if ( access(TSEQU_CONFIG_FILENAME,R_OK) != 0 ) {
		msgbox_error(	NULL,
				msgbox_string1 ("Failed to open '%s' for reading! ( Permission Denied ).",TSEQU_CONFIG_FILENAME),
				msgbox_string2 ("Checkout TSUSB Driver Installation."),
				GTK_BUTTONS_OK
				);
		return 4;
	}
	if ( access(TSEQU_CONFIG_FILENAME,W_OK) != 0 ) {
		msgbox_error(	NULL,
				msgbox_string1 ("Failed to open '%s' for writing! ( Permission Denied ).",TSEQU_CONFIG_FILENAME),
				msgbox_string2 ("Checkout TSUSB Driver Installation."),
				GTK_BUTTONS_OK
				);
		return 5;
	}

	if (!options_load_configuration()) {
		msgbox_error(	NULL,
				msgbox_string1 ("Failed to load config from '%s' file.",TSEQU_CONFIG_FILENAME),
				msgbox_string2 ("Checkout TSUSB Driver Installation."),
				GTK_BUTTONS_OK
				);
		return 6;
	}

	window_main = create_window_main ();

	gtk_window_set_title (GTK_WINDOW (window_main), _(TSEQU_PROGRAM_TITLE " V" TSEQU_PROGRAM_VERSION));

	device_treeview_create(window_main);

	window_options = create_window_options ();

	gtk_window_set_title (GTK_WINDOW (window_options), _(TSEQU_PROGRAM_TITLE " V" TSEQU_PROGRAM_VERSION));

	options_initialize ( window_main, window_options );

	window_calibrate = create_window_calibrate();

	set_label_string (lookup_widget (window_calibrate,  "label_version"),
					TRUE,
					"<span foreground=\"#000000\">"
					"Version " TSEQU_PROGRAM_VERSION
					"</span>"
					);

	screen_width  = gdk_screen_width();
	screen_height = gdk_screen_height();

	set_label_string (lookup_widget (window_calibrate,  "label_screen"),
					TRUE,
					"<span foreground=\"#000000\">"
					"Screen resolution %d x %d"
					"</span>",
					screen_width,
					screen_height);

	gtk_widget_set_size_request (window_calibrate, screen_width, screen_height);

	gtk_container_set_border_width (GTK_CONTAINER (window_calibrate), TS_CALIB_BOX_BORDER);

	calibrate_initialize ( window_main, window_calibrate );

	gtk_widget_show (window_main);

	tsevdev_pool_start();

	pcsound_open();

	gtk_main ();

	pcsound_close();

	tsevdev_pool_end();

	tsevdev_close();

	return 0;

}

