/*
 * msgbox.c
 *
 */

#include <gtk/gtk.h>

#include "pcsound.h"

#define MB_STRING_SIZE 255

gchar mb_string1[MB_STRING_SIZE];
gchar mb_string2[MB_STRING_SIZE];

gchar *
msgbox_string1 ( const gchar *message_format, ... )
{
	va_list args;

	if ( !message_format ) return NULL;

	va_start (args, message_format);
	vsnprintf ( mb_string1, MB_STRING_SIZE, message_format, args);
	va_end (args);

	return mb_string1;
}

gchar *
msgbox_string2 ( const gchar *message_format, ... )
{
	va_list args;

	if ( !message_format ) return NULL;

	va_start (args, message_format);
	vsnprintf ( mb_string2, MB_STRING_SIZE, message_format, args);
	va_end (args);

	return mb_string2;
}

gint
msgbox (
	GtkWidget *window,
	GtkDialogFlags flags,
	GtkMessageType type,
	GtkButtonsType buttons,
	const gchar *msg_string1,
	const gchar *msg_string2
	)
{
	gint ret;

	GtkWidget *dialog;

	dialog = gtk_message_dialog_new (
			GTK_WINDOW (window),
			flags,
			type,
			buttons,
			"%s\n%s",msg_string1,msg_string2);
	gtk_window_set_position (GTK_WINDOW (dialog), GTK_WIN_POS_CENTER);

	ret = gtk_dialog_run (GTK_DIALOG (dialog));

	gtk_widget_destroy (dialog);

	return ret;

}

GtkButtonsType
msgbox_buttons ( GtkButtonsType buttons )
{
	if ( buttons <  GTK_BUTTONS_NONE ) return  GTK_BUTTONS_OK;

	if ( buttons > GTK_BUTTONS_OK_CANCEL ) return GTK_BUTTONS_OK;

	return buttons;
}

gint
msgbox_info ( GtkWidget *window, const gchar *msg_string1, const gchar *msg_string2, GtkButtonsType buttons )
{
	return	msgbox ( window,
			GTK_DIALOG_MODAL | GTK_DIALOG_DESTROY_WITH_PARENT,
			GTK_MESSAGE_INFO,
			msgbox_buttons(buttons),
			msg_string1,
			msg_string2
			);
}

gint
msgbox_warning ( GtkWidget *window, const gchar *msg_string1, const gchar *msg_string2, GtkButtonsType buttons )
{
	return	msgbox ( window,
			GTK_DIALOG_MODAL | GTK_DIALOG_DESTROY_WITH_PARENT,
			GTK_MESSAGE_WARNING,
			msgbox_buttons(buttons),
			msg_string1,
			msg_string2
			);
}

gint
msgbox_question ( GtkWidget *window, const gchar *msg_string1, const gchar *msg_string2, GtkButtonsType buttons )
{
	return	msgbox ( window,
			GTK_DIALOG_MODAL | GTK_DIALOG_DESTROY_WITH_PARENT,
			GTK_MESSAGE_QUESTION,
			msgbox_buttons(buttons),
			msg_string1,
			msg_string2
			);
}

gint
msgbox_error ( GtkWidget *window, const gchar *msg_string1, const gchar *msg_string2, GtkButtonsType buttons )
{
	pcsound_onetime(100,3000,50);
	return	msgbox ( window,
			GTK_DIALOG_MODAL | GTK_DIALOG_DESTROY_WITH_PARENT,
			GTK_MESSAGE_ERROR,
			msgbox_buttons(buttons),
			msg_string1,
			msg_string2
			);
}

