/*
 * options.c
 *
 */

#include <gtk/gtk.h>

#include <stdio.h>
#include <string.h>

#include "support.h"
#include "tsequ.h"
#include "tsusbhid.h"
#include "tsevdev.h"
#include "tshiddev.h"

enum {
	VB_UNUSE,
	VB_STREAM,
	VB_LIFT_OFF,
	VB_TOUCH_DOWN,
	VB_DRAG_DROP
	};
enum {
	FS_UNUSE,
	FS_SLIENT,
	FS_LIFT_OFF_ONLY,
	FS_TOUCH_DOWN_ONLY,
	FS_BOTH_TOUCH_DOWN_AND_LIFT_OFF
	};
enum {
	OP_DISABLED,
	OP_ENABLED
	};

#define	OPTION_VB_MODE_MIN		VB_STREAM
#define	OPTION_VB_MODE_DEFAULT		VB_DRAG_DROP
#define	OPTION_VB_MODE_MAX		VB_DRAG_DROP

#define	OPTION_RB_AUTO_MODE_MIN		OP_DISABLED
#define	OPTION_RB_AUTO_MODE_DEFAULT	OP_ENABLED
#define	OPTION_RB_AUTO_MODE_MAX		OP_ENABLED

#define	OPTION_RB_AUTO_STIME_MIN	0
#define	OPTION_RB_AUTO_STIME_DEFAULT	1
#define	OPTION_RB_AUTO_STIME_MAX	9

#define	OPTION_RB_AUTO_ETIME_MIN	0
#define	OPTION_RB_AUTO_ETIME_DEFAULT	1
#define	OPTION_RB_AUTO_ETIME_MAX	9

#define	OPTION_RB_AUTO_BEEP_MIN		OP_DISABLED
#define	OPTION_RB_AUTO_BEEP_DEFAULT	OP_ENABLED
#define	OPTION_RB_AUTO_BEEP_MAX		OP_ENABLED

#define	OPTION_FS_MODE_MIN		FS_SLIENT
#define	OPTION_FS_MODE_DEFAULT		FS_BOTH_TOUCH_DOWN_AND_LIFT_OFF
#define	OPTION_FS_MODE_MAX		FS_BOTH_TOUCH_DOWN_AND_LIFT_OFF

#define	OPTION_FS_TONE_MIN		200
#define	OPTION_FS_TONE_DEFAULT		2468
#define	OPTION_FS_TONE_MAX		5000

#define	OPTION_FS_DURATION_MIN		10
#define	OPTION_FS_DURATION_DEFAULT	16
#define	OPTION_FS_DURATION_MAX		50

#define	OPTION_CP_MODE_MIN		TSEQU_CALIBRATION_4POINT_MODE1
#define	OPTION_CP_MODE_DEFAULT		TSEQU_CALIBRATION_4POINT_MODE2
#define	OPTION_CP_MODE_MAX		TSEQU_CALIBRATION_4POINT_MODE2

static GtkWidget *window_main;
static GtkWidget *window_options;

static gint use_devnum = -1;

static gint use_calibration_mode = OPTION_CP_MODE_DEFAULT;
static gint new_calibration_mode = OPTION_CP_MODE_DEFAULT;

static struct tsusb_feature_report use_feature;
static struct tsusb_feature_report new_feature;

gint options_now_calibration_mode(void)
{
	return use_calibration_mode;
}

gboolean options_load_configuration(void)
{

	FILE *fp;

	const gchar* szCalibrationMode = "CalibrationMode";

	gchar s[255], t[255], n[255];

	gint i, j, k, value;
	gint slen;

	fp = fopen(TSEQU_CONFIG_FILENAME,"r");

	if ( fp == NULL ) {
		return FALSE;
	}

	while ( (fgets(s,255,fp)) != NULL )
	{

		if ( s[0] == '#'  ) continue;
		if ( s[0] == '\n' ) continue;

		slen = strlen(s);
		if ( slen < 5 ) continue;

		for ( i=0,j=0;i<slen;i++) {
			if( j>0 || ( j==0 && !isspace(s[i]) ) ) { t[j]=s[i]; j++; }
			if ( s[i] == '=' ) break;
		}
		t[j] = '\0';

		if ( j==0 || i==slen ) continue;

		for ( i=j,j=0; i<slen; i++ ) {
			if ( s[i] == '\n' || s[i] == '\0' ) break;
			n[j] = s[i];
			j++;
		}
		n[j] = '\0';
		value = atoi(n);

		if ( strncasecmp(t,szCalibrationMode,strlen(szCalibrationMode)) == 0 ){
			use_calibration_mode = value;
			continue;
		}

	}

	fclose(fp);

	if (( use_calibration_mode != TSEQU_CALIBRATION_4POINT_MODE1 ) &&
	    ( use_calibration_mode != TSEQU_CALIBRATION_4POINT_MODE2 ) ) {
	      use_calibration_mode  = TSEQU_CALIBRATION_4POINT_MODE2;
	}

	return TRUE;

}

gboolean options_save_configuration()
{

	FILE *fp;

	fp = fopen(TSEQU_CONFIG_FILENAME,"w+");

	if ( fp == NULL ) {
		return FALSE;
	}

	fprintf( fp, "# *****************************************************************\n");
	fprintf( fp, "# %s ... %s Configuration File \n",TSEQU_CONFIG_FILENAME,TSEQU_CONFIG_TITLE);
	fprintf( fp, "# *****************************************************************\n");
	fprintf( fp, "# WARNING! Incorrect changes to this file could prevent the\n");
	fprintf( fp, "#          Touchscreen from operating or operating correctly!\n");
	fprintf( fp, "# *****************************************************************\n");
	fprintf( fp, "\n");
	fprintf( fp, "CalibrationMode=%d\n",use_calibration_mode);

	fclose(fp);

	return TRUE;

}

static void
set_combo_box_entry_editable (const gchar *widget_name, gboolean editable)
{
	GtkComboBoxEntry *entry_box = GTK_COMBO_BOX_ENTRY(lookup_widget (window_options,widget_name ));
	gtk_entry_set_editable( GTK_ENTRY (GTK_BIN(entry_box)->child), editable );
}

static void
set_combo_box_entry_active (const gchar *widget_name, gint index )
{
	GtkWidget *entry_box = lookup_widget (window_options,widget_name );
	gtk_combo_box_set_active( GTK_COMBO_BOX(entry_box), index );
}

static gint
get_combo_box_entry_active (const gchar *widget_name)
{
	GtkWidget *entry_box = lookup_widget (window_options,widget_name );
	return gtk_combo_box_get_active( GTK_COMBO_BOX(entry_box) );
}

static void
set_spin_button_value (const gchar *widget_name, gdouble value )
{
	GtkWidget *spin_button = lookup_widget (window_options,widget_name );
	gtk_spin_button_set_value( GTK_SPIN_BUTTON(spin_button), value );
}

static gdouble
get_spin_button_value (const gchar *widget_name)
{
	GtkWidget *spin_button = lookup_widget (window_options,widget_name );
	return gtk_spin_button_get_value( GTK_SPIN_BUTTON(spin_button) );
}

static void
options_set_VB_mode(gint virtual_button)
{
	set_combo_box_entry_active(
		"comboboxentry_VB_mode",
		virtual_button - OPTION_VB_MODE_MIN
		);
}

static gint
options_get_VB_mode(void)
{
	return get_combo_box_entry_active("comboboxentry_VB_mode") + OPTION_VB_MODE_MIN;
}

static void
options_set_RB_auto_mode(gint right_button_auto_mode)
{
	set_combo_box_entry_active(
		"comboboxentry_RB_auto_mode",
		right_button_auto_mode - OPTION_RB_AUTO_MODE_MIN
		);
}

static gint
options_get_RB_auto_mode(void)
{
	return get_combo_box_entry_active("comboboxentry_RB_auto_mode") + OPTION_RB_AUTO_MODE_MIN;
}

static void
options_set_RB_auto_stime(gint right_button_auto_stime)
{
	set_combo_box_entry_active(
		"comboboxentry_RB_auto_stime",
		right_button_auto_stime - OPTION_RB_AUTO_STIME_MIN
		);
}

static gint
options_get_RB_auto_stime(void)
{
	return get_combo_box_entry_active("comboboxentry_RB_auto_stime") + OPTION_RB_AUTO_STIME_MIN;
}

static void
options_set_RB_auto_etime(gint right_button_auto_etime)
{
	set_combo_box_entry_active(
		"comboboxentry_RB_auto_etime",
		right_button_auto_etime - OPTION_RB_AUTO_ETIME_MIN
		);
}

static gint
options_get_RB_auto_etime(void)
{
	return get_combo_box_entry_active("comboboxentry_RB_auto_etime") + OPTION_RB_AUTO_ETIME_MIN;
}

static void
options_set_RB_auto_beep(gint right_button_auto_beep)
{
	set_combo_box_entry_active(
		"comboboxentry_RB_auto_beep",
		right_button_auto_beep - OPTION_RB_AUTO_BEEP_MIN
		);
}

static gint
options_get_RB_auto_beep(void)
{
	return get_combo_box_entry_active("comboboxentry_RB_auto_beep") + OPTION_RB_AUTO_BEEP_MIN;

}

static void
options_set_FS_mode(gint feedback_sound)
{
	set_combo_box_entry_active(
		"comboboxentry_FS_mode",
		feedback_sound - OPTION_FS_MODE_MIN
		);
}

static gint
options_get_FS_mode(void)
{
	return get_combo_box_entry_active("comboboxentry_FS_mode") + OPTION_FS_MODE_MIN;
}

static void
options_set_FS_tone(gint feedback_tone)
{
	set_spin_button_value(
		"spinbutton_FS_tone",
		feedback_tone
		);
}

static gint
options_get_FS_tone(void)
{
	return get_spin_button_value ("spinbutton_FS_tone");
}

static void
options_set_FS_duration(gint feedback_duration)
{
	set_spin_button_value(
		"spinbutton_FS_duration",
		feedback_duration
		);
}

static gint
options_get_FS_duration(void)
{
	return get_spin_button_value ("spinbutton_FS_duration");
}

static void
options_set_CP_mode(gint calibration_mode)
{
	set_combo_box_entry_active(
		"comboboxentry_CP_mode",
		calibration_mode - OPTION_CP_MODE_MIN
		);
}

static gint
options_get_CP_mode(void)
{
	return get_combo_box_entry_active("comboboxentry_CP_mode") + OPTION_CP_MODE_MIN;
}

static gboolean
options_button_state()
{
	gboolean flag = TRUE;

	if ( use_calibration_mode == new_calibration_mode ) {
        	if ( memcmp(&use_feature,&new_feature,sizeof(struct tsusb_feature_report)) == 0 ) {
			flag = FALSE;
		}
	}

	gtk_widget_set_sensitive( lookup_widget(window_options,"button_undo" ), flag );
	gtk_widget_set_sensitive( lookup_widget(window_options,"button_apply"), flag );

	return flag;
}

static void
options_show_settings(struct tsusb_feature_report *pr)
{

	if ( pr->virtual_button < OPTION_VB_MODE_MIN ||
	     pr->virtual_button > OPTION_VB_MODE_MAX ){
	     pr->virtual_button = OPTION_VB_MODE_DEFAULT;
	}
	options_set_VB_mode(pr->virtual_button);

	if ( pr->right_button_auto_mode > OPTION_RB_AUTO_MODE_MAX ){
	     pr->right_button_auto_mode = OPTION_RB_AUTO_MODE_DEFAULT;
	}
	options_set_RB_auto_mode(pr->right_button_auto_mode);

	pr->right_button_function = pr->right_button_auto_mode;

	if ( pr->right_button_auto_stime > OPTION_RB_AUTO_STIME_MAX ){
	     pr->right_button_auto_stime = OPTION_RB_AUTO_STIME_DEFAULT;
	}
	options_set_RB_auto_stime(pr->right_button_auto_stime);

	if ( pr->right_button_auto_etime > OPTION_RB_AUTO_ETIME_MAX ){
	     pr->right_button_auto_etime = OPTION_RB_AUTO_ETIME_DEFAULT;
	}
	options_set_RB_auto_etime(pr->right_button_auto_etime);

	if ( pr->right_button_auto_beep > OPTION_RB_AUTO_BEEP_MAX ){
	     pr->right_button_auto_beep = OPTION_RB_AUTO_BEEP_DEFAULT;
	}
	options_set_RB_auto_beep(pr->right_button_auto_beep);

	if ( pr->feedback_sound < OPTION_FS_MODE_MIN ||
	     pr->feedback_sound > OPTION_FS_MODE_MAX ){
	     pr->feedback_sound = OPTION_FS_MODE_DEFAULT;
	}
	options_set_FS_mode(pr->feedback_sound);

	if ( pr->feedback_tone < OPTION_FS_TONE_MIN ||
	     pr->feedback_tone > OPTION_FS_TONE_MAX ){
	     pr->feedback_tone = OPTION_FS_TONE_DEFAULT;
	}
	options_set_FS_tone(pr->feedback_tone);

	if ( pr->feedback_duration < OPTION_FS_DURATION_MIN ||
	     pr->feedback_duration > OPTION_FS_DURATION_MAX ){
	     pr->feedback_duration = OPTION_FS_DURATION_DEFAULT;
	}
	options_set_FS_duration(pr->feedback_duration);

	options_set_CP_mode(new_calibration_mode);

}

static void
options_init_settings()
{
	memcpy(&new_feature,&use_feature,sizeof(struct tsusb_feature_report));
	new_calibration_mode = use_calibration_mode;
	options_show_settings(&new_feature);
	options_button_state();
}

void options_VB_mode_changed(void)
{
	new_feature.virtual_button = options_get_VB_mode();
	options_button_state();
}

void options_RB_auto_mode_changed(void)
{
	new_feature.right_button_auto_mode = options_get_RB_auto_mode();
	new_feature.right_button_function = new_feature.right_button_auto_mode;

	options_button_state();
}

void options_RB_auto_stime_changed(void)
{
	new_feature.right_button_auto_stime = options_get_RB_auto_stime();
	options_button_state();
}

void options_RB_auto_etime_changed(void)
{
	new_feature.right_button_auto_etime = options_get_RB_auto_etime();
	options_button_state();
}

void options_RB_auto_beep_changed(void)
{
	new_feature.right_button_auto_beep = options_get_RB_auto_beep();
	options_button_state();
}

void options_FS_mode_changed(void)
{
	new_feature.feedback_sound = options_get_FS_mode();
	options_button_state();
}

void options_FS_tone_changed(void)
{
	new_feature.feedback_tone = options_get_FS_tone();
	options_button_state();
}

void options_FS_duration_changed(void)
{
	new_feature.feedback_duration = options_get_FS_duration();
	options_button_state();
}

void options_CP_mode_changed(void)
{
	new_calibration_mode = options_get_CP_mode();
	options_button_state();
}

void
options_start(void)
{
	gint ret;

	use_devnum = device_treeview_selected_devnum();

	ret = tshiddev_open(use_devnum);
	if ( ret < 0 ) {
		msgbox_error ( window_main,
		msgbox_string1 ( "options_start: failed to open device" ),
		msgbox_string2 ( "device '" TSEVDEV_DEVICE_NAME "%d', error %d", use_devnum, ret )
		);
		return;
	}

	ret = tshiddev_get_feature(&use_feature);
	if ( ret < 0 ) {
		msgbox_error ( window_main,
		msgbox_string1 ( "options_start: failed to get feature from device" ),
		msgbox_string2 ( "device '" TSEVDEV_DEVICE_NAME "%d', error %d", use_devnum, ret )
		);
		tshiddev_close();
		return;
	}

	options_init_settings();

	gtk_widget_hide (window_main);
	gtk_widget_show (window_options);

}

void
options_end(void)
{
	tshiddev_close();

	gtk_widget_hide (window_options);
	gtk_widget_show (window_main);
}

void
options_initialize ( GtkWidget *widget_main, GtkWidget *widget_options )
{
	window_main = widget_main;
	window_options = widget_options;

	set_combo_box_entry_editable("comboboxentry_VB_mode"      ,FALSE);
	set_combo_box_entry_editable("comboboxentry_RB_auto_mode" ,FALSE);
	set_combo_box_entry_editable("comboboxentry_RB_auto_stime",FALSE);
	set_combo_box_entry_editable("comboboxentry_RB_auto_etime",FALSE);
	set_combo_box_entry_editable("comboboxentry_RB_auto_mode" ,FALSE);
	set_combo_box_entry_editable("comboboxentry_RB_auto_beep" ,FALSE);
	set_combo_box_entry_editable("comboboxentry_FS_mode"      ,FALSE);
	set_combo_box_entry_editable("comboboxentry_CP_mode"      ,FALSE);

}


void
options_button_default_clicked(void)
{

	new_feature.virtual_button		= OPTION_VB_MODE_DEFAULT;
	new_feature.right_button_auto_mode	= OPTION_RB_AUTO_MODE_DEFAULT;
	new_feature.right_button_function       = new_feature.right_button_auto_mode;
	new_feature.right_button_auto_stime	= OPTION_RB_AUTO_STIME_DEFAULT;
	new_feature.right_button_auto_etime	= OPTION_RB_AUTO_ETIME_DEFAULT;
	new_feature.right_button_auto_beep	= OPTION_RB_AUTO_BEEP_DEFAULT;
	new_feature.feedback_sound		= OPTION_FS_MODE_DEFAULT;
	new_feature.feedback_tone		= OPTION_FS_TONE_DEFAULT;
	new_feature.feedback_duration		= OPTION_FS_DURATION_DEFAULT;

	new_calibration_mode = OPTION_CP_MODE_DEFAULT;

	options_show_settings(&new_feature);

}


void
options_button_undo_clicked(void)
{
	options_init_settings();
	options_button_state();
}

void
options_button_apply_clicked(void)
{
	gint ret;
	gint tmp_calibration_mode;

	use_calibration_mode = new_calibration_mode;

	if (!options_save_configuration() ) {
		msgbox_error ( window_options,
			msgbox_string1 ("Failed to save config!"),
			msgbox_string2 ("File: '%s'",TSEQU_CONFIG_FILENAME)
			);
        	use_calibration_mode = tmp_calibration_mode;
		return;
	}

	new_feature.feature_id = TSUSB_FEATURE_ID_SAVE;

	ret = tshiddev_set_feature(&new_feature);

	if ( ret < 0 ) {
		msgbox_error ( window_options,
			msgbox_string1 ( "Failed to save new options to device!" ),
			msgbox_string2 ( "device '" TSEVDEV_DEVICE_NAME "%d', error %d", use_devnum, ret )
			);
		return;
	}

        memcpy(&use_feature,&new_feature,sizeof(struct tsusb_feature_report));

	ret = tshiddev_disable();

	if ( ret < 0 ) {
		msgbox_error ( window_options,
			msgbox_string1 ( "Failed to disable the device!" ),
			msgbox_string2 ( "device '" TSEVDEV_DEVICE_NAME "%d', error %d", use_devnum, ret )
			);
		return;
	}

	ret = tshiddev_enable();

	if ( ret < 0 ) {
		msgbox_error ( window_options,
			msgbox_string1 ( "Failed to auto enable the device!" ),
			msgbox_string2 ( "device '" TSEVDEV_DEVICE_NAME "%d', error %d", use_devnum, ret )
			);
		return;
	}

	options_button_state();

}

void
options_button_done_clicked(void)
{
	if (options_button_state()) {
		options_button_apply_clicked();
	}
	options_end();
}

