/*
 * pcsound.c
 *
 */

#include <gtk/gtk.h>

#include <fcntl.h>
#include <linux/kd.h>

#include "pcsound.h"

#define PCSOUND_DEVICE_NAME "/dev/console"

static gint	pcsound_fd = -1;
static gint	pcsound_proc_timer = 0;
static gint	pcsound_table_index = 0;

void
pcsound_open (void)
{
	pcsound_fd = open(PCSOUND_DEVICE_NAME, O_RDWR );
	if ( pcsound_fd < 0 ) {
		return;
	}
}

void
pcsound_close (void)
{
	close(pcsound_fd);
}

void
pcsound_output (int loudness, int pitch, int duration)
{
	if (loudness && pitch) {
		ioctl(pcsound_fd, KDMKTONE,
		      ((1193190 / pitch) & 0xffff) |
		      (((unsigned long)duration *
			loudness / 50) << 16));
	}
}

static gint
pcsound_proc (gpointer data)
{
	pcsound_table	*pcsound_table_data = (pcsound_table *)data;

	if ( pcsound_table_data[pcsound_table_index].loudness ) {
		pcsound_output(
			pcsound_table_data[pcsound_table_index].loudness,
			pcsound_table_data[pcsound_table_index].pitch,
			pcsound_table_data[pcsound_table_index].duration
			);
		pcsound_table_index++;
		return 1;
	}

	if ( pcsound_proc_timer ) {
		g_source_remove(pcsound_proc_timer);
		pcsound_proc_timer = 0;
	}

	return 0;
}

void
pcsound_proc_start(pcsound_table *pcsound_table_data)
{

	if ( pcsound_fd < 0 ) {
		return;
	}

	pcsound_table_index = 0;
	pcsound_proc_timer = g_timeout_add(
					50,
					pcsound_proc,
					pcsound_table_data
					);

}

void
pcsound_onetime (int loudness, int pitch, int duration)
{
	gint fd = -1;

	fd = open(PCSOUND_DEVICE_NAME, O_RDWR );

	if ( fd < 0 ) {
		return;
	}

	if (loudness && pitch) {
		ioctl(fd, KDMKTONE,
		      ((1193190 / pitch) & 0xffff) |
		      (((unsigned long)duration *
			loudness / 50) << 16));
	}

	close(fd);
}

void
pcsound_test(void)
{
	static pcsound_table pcsound_test_data [] = {
	{100, 1600,50},
	{100, 1300,50},
	{100, 1000,50},
	{0},
	};
	pcsound_proc_start(pcsound_test_data);
}

