/*
 * tsevdev.c
 *
 */

#include <gtk/gtk.h>

#include <stdio.h>
#include <string.h>
#include <fcntl.h>
#include <sys/poll.h>
#include <errno.h>

#include "support.h"

#include "tsusbhid.h"
#include "tsevdev.h"

#define	TSEVDEV_POLL_CLOCK	500	/* millseconds	*/
#define	TSEVDEV_BUFFER_SIZE	256	/* size of reception buffer */

enum {
	COL_FLAG,
	COL_NAME,
	COL_DESC,
	COL_DEVNUM,
	NUM_COLS };

static GtkWidget *window_main;
static GtkWidget *device_treeview;

static gint		tsevdev_fd = -1;
static struct pollfd	tsevdev_udfs;
static int 		tsevdev_bufpos = 0;
static unsigned char	tsevdev_buffer[TSEVDEV_BUFFER_SIZE];
static gint		tsevdev_poll_timer = 0;

static void
device_pluged(gint devnum)
{
	GtkTreeModel	*model;
	GtkListStore	*store;
	GtkTreeIter	iter;
	gboolean	valid;
	gint		num;
	gchar		name[10];
	gchar		desc[50];

	model = gtk_tree_view_get_model(GTK_TREE_VIEW (device_treeview));

	store = GTK_LIST_STORE(model);

	valid = gtk_tree_model_get_iter_first(model,&iter);
	while (valid) {
		gtk_tree_model_get(model,&iter,COL_DEVNUM,&num,-1);
		if ( num >= devnum ) {
			/* NOTE: num = devnum is never happen, otherwise there are something wrong in our code */
			if ( num > devnum ) {
				gtk_list_store_insert_before(store,&iter,&iter);
			}
			break;
		}
		valid = gtk_tree_model_iter_next(model,&iter);
	}
	if (!valid) gtk_list_store_append(store,&iter);

	snprintf(name,sizeof(name),"tsusb#%d",devnum+1);
	tshiddev_get_id_string(devnum,desc, sizeof(desc));
	gtk_list_store_set (store, &iter,
				COL_FLAG, "",
				COL_NAME, name,
				COL_DEVNUM, devnum,
				COL_DESC, desc,
				-1);
}

static void
device_unpluged(gint devnum)
{
	GtkTreeModel	*model;
	GtkListStore	*store;
	GtkTreeIter	iter;
	gboolean		valid;
	gint			num;

	model = gtk_tree_view_get_model(GTK_TREE_VIEW (device_treeview));

	store = GTK_LIST_STORE(model);

	valid = gtk_tree_model_get_iter_first(model,&iter);
	while (valid) {
		gtk_tree_model_get(model,&iter,COL_DEVNUM,&num,-1);
		if ( num == devnum ) {
			gtk_list_store_remove(store,&iter);
			break;
		}
		valid = gtk_tree_model_iter_next(model,&iter);
	}
}

gint
device_treeview_selected_devnum (void)
{
	GtkTreeSelection *selection;
	GtkTreeModel *model;
	GtkTreeIter iter;
	gint devnum = -1;

	/* NOTE: This will only work in single or browse selection mode! */
	selection = gtk_tree_view_get_selection(GTK_TREE_VIEW (device_treeview));

	if ( gtk_tree_selection_get_selected(selection, &model, &iter) ) {
		gtk_tree_model_get(model, &iter, COL_DEVNUM, &devnum, -1);
	}

	return devnum;
}

static gboolean
device_treeview_selection_func (GtkTreeSelection *selection,
						 GtkTreeModel *model,
						 GtkTreePath *path,
						 gboolean path_currently_selected,
						 gpointer userdata)
{
	GtkListStore *store = GTK_LIST_STORE(model);
	GtkTreeIter iter;
	gint devnum;
	if ( gtk_tree_model_get_iter(model,&iter,path) ) {
		gtk_tree_model_get(model, &iter, COL_DEVNUM, &devnum, -1);
		if ( !path_currently_selected ) {
			gtk_list_store_set (store, &iter, COL_FLAG, "*",-1);
			gtk_widget_set_sensitive( lookup_widget(window_main,"button_calibration"), TRUE );
			gtk_widget_set_sensitive( lookup_widget(window_main,"button_options"    ), TRUE );
		}
		else {
			gtk_list_store_set (store, &iter, COL_FLAG, "",-1);
			gtk_widget_set_sensitive( lookup_widget(window_main,"button_calibration"), FALSE );
			gtk_widget_set_sensitive( lookup_widget(window_main,"button_options"    ), FALSE );
		}
	}
	return TRUE;
}

void
device_treeview_create (GtkWidget *window)
{

	GtkCellRenderer	*renderer;
	GtkTreeModel		*model;
	GtkListStore		*store;
	GtkTreeIter		iter;

	window_main = window;
	device_treeview = lookup_widget (window_main, "treeview_devices");

	renderer = gtk_cell_renderer_text_new();
	gtk_tree_view_insert_column_with_attributes(GTK_TREE_VIEW (device_treeview),
										-1,
										"",
										renderer,
										"text",
										COL_FLAG,
										NULL
										);

	renderer = gtk_cell_renderer_text_new();
	gtk_tree_view_insert_column_with_attributes(GTK_TREE_VIEW (device_treeview),
										-1,
										"Device",
										renderer,
										"text",
										COL_NAME,
										NULL
										);

	renderer = gtk_cell_renderer_text_new();
	gtk_tree_view_insert_column_with_attributes(GTK_TREE_VIEW (device_treeview),
										-1,
										"Description",
										renderer,
										"text",
										COL_DESC,
										NULL
										);

	gtk_tree_view_set_headers_visible (GTK_TREE_VIEW (device_treeview), TRUE);

	store = gtk_list_store_new( NUM_COLS, G_TYPE_STRING, G_TYPE_STRING,  G_TYPE_STRING, G_TYPE_UINT );

	model = GTK_TREE_MODEL(store);

	gtk_tree_view_set_model(GTK_TREE_VIEW (device_treeview), model);

	g_object_unref(model);

	GtkTreeSelection *selection;

	selection = gtk_tree_view_get_selection(GTK_TREE_VIEW (device_treeview));

	gtk_tree_selection_set_select_function(selection,device_treeview_selection_func,NULL,NULL);

	return;
}

gint
tsevdev_open (void)
{
	gint ret;

	ret = open (TSEVDEV_DEVICE_NAME, O_RDWR );
	if ( ret < 0 ) {
		printf("tsevdev_open: Failed to open '%s' device!  (ret=%d)\n",TSEVDEV_DEVICE_NAME, ret);
		return FALSE;
	}

	tsevdev_fd = ret;

	tsevdev_udfs.fd     = tsevdev_fd;
	tsevdev_udfs.events = POLLIN;

	tsevdev_bufpos = 0;

	return TRUE;
}

void
tsevdev_close (void)
{
	close(tsevdev_fd);
}

static void
tsevdev_proc_event ( struct tsusb_event_packet *event )
{
	switch (event->id)
	{
		case TSUSB_EVENT_PLUG:
			device_pluged(event->v1);
			break;
		case TSUSB_EVENT_UNPLUG:
			device_unpluged(event->v1);
			break;
		case TSUSB_EVENT_INPUT_DATA:
			break;
		case TSUSB_EVENT_ENABLE:
			break;
		case TSUSB_EVENT_DISABLE:
			break;
		default:
			break;
	}
}

static void
tsevdev_read (void)
{

	int packetLength = sizeof(struct tsusb_event_packet);

	int len, pos, remaining;

	remaining = sizeof(tsevdev_buffer) - tsevdev_bufpos;

	len = read(tsevdev_fd, tsevdev_buffer+tsevdev_bufpos, remaining);

	if (len <= 0) return;

	tsevdev_bufpos += len;

	pos = 0;

	while ((tsevdev_bufpos - pos) >=  packetLength)
	{
		tsevdev_proc_event( (struct tsusb_event_packet *)(tsevdev_buffer + pos) );
		pos += packetLength;
	}

	if (pos) {
		if (pos < tsevdev_bufpos)
		{
			memmove(tsevdev_buffer,tsevdev_buffer+pos,tsevdev_bufpos-pos);
			tsevdev_bufpos -= pos;
		}
		else
		{
			tsevdev_bufpos = 0;
		}
	}

}

static gint
tsevdev_pool (gpointer data)
{
	gint ret;

	ret = poll(&tsevdev_udfs,1,0);

	if ( ret > 0 ) tsevdev_read();

	return 1;
}

void
tsevdev_pool_start (void)
{
	tsevdev_poll_timer = g_timeout_add(
				TSEVDEV_POLL_CLOCK,
				tsevdev_pool,
				NULL
				);
}

void
tsevdev_pool_end (void)
{
	if (tsevdev_poll_timer) {
		g_source_remove(tsevdev_poll_timer);
		tsevdev_poll_timer = 0;
	}
}

